function obj = enable_linesearch(obj, varargin)
% ENABLE_LINESEARCH - Turn on Armijo line search (sufficient decrease criteria)
%
% Given a starting point x_k, a step p_k, and a real-valued function f the
% Armijo line search criterion reduces the step length (initially eta = 1)
% till f(x_k + eta * p_k) - f(x_k) <= c * eta * p_k * f'(x_k)
%
% Arguments:
%
% obj: An instance of @glmparams
% c: Stringency of the Armijo test (c << 1) [opt: defaults to 1e-3]
% rho: Factor to reduce the step size by [opt: default to 1/2]
% min_armijo_iter: Number of iterations to wait before using Armijo [opt:
%                  default to -1, ie -- always use Armijo test].
% mineta: Minimum value of eta to try before giving on the update [opt:
%         default to rho^5]

% Author: ajit@cs.cmu.edu (Ajit Singh)

[c, rho, min_armijo_iter, mineta] = ...
    process_options(varargin, 'c', 1e-3, 'rho', 0.5, 'min_armijo_iter', -1, ...
                    'mineta', 0.5^5);

assert(isbetween(c, 0.0, 1.0, 'open'));
if c > 0.01
  warning('c should typically be << 0.01, do you really want c = %g ?', c);
end

assert(isbetween(rho, 0.0, 1.0, 'open'), ...
       'rho must be a scaling factor in (0,1) (is %g)', rho);
assert(fix(min_armijo_iter) == min_armijo_iter);
assert(isfinite(min_armijo_iter));
assert(mineta > 0.0);
assert(mineta < rho, 'mineta >= rho (%g vs. %g) forbids any Armijo steps', ...
       mineta, rho);

obj.linesearch = true;
obj.c = c;
obj.rho = rho;
obj.min_armijo_iter = min_armijo_iter;
obj.mineta = mineta;
