function [luv lvz lvr] = evaluate_loss(model, split, params, varargin)
% EVALUTE_LOSS - Computes the loss on held out data from X and Y.
%
% Arguments:
%
% model: An instance of @model3f, containing factors U, V, and Z.
% split: A split of the data, containing fields xtest, exi, exj, ytest, eyi, eyj
% params: Optimization parameters. f/loss_f and ff/loss_ff define the link
%         and loss.
% mode: Controls whether we compute the loss on X, Y, or both ('all'). Opt arg.
%
% Return Values:
%
% luv: Loss on test data from X.
% lvz: Loss of test data from Y.

% Author: Ajit Singh (ajit@cs.cmu.edu)

mode = process_options(varargin, 'mode', 'all');

if strcmpi(mode, 'X')
  luv = xloss(model, split, params);
  lvz = NaN;
  lvr = NaN;
elseif strcmpi(mode, 'Y')
  luv = NaN;
  lvz = yloss(model, split, params);
  lvr = NaN;
elseif strcmpi(mode, 'all')
  luv = xloss(model, split, params);
  lvz = yloss(model, split, params);
  lvr = qloss(model, split, params);
else
  error('Parameter mode = %s is invalid', mode);
end

function val = xloss(model, split, params)
% XLOSS - Compute the loss on held out test examples from X

% TODO(ajit): What if the weights were actually the data weights ?
x = split.xtest;
w = repmat(1/length(x), 1, length(x));
U = model.U;
V = model.V;
val = evaluate(x, w, split.exi, split.exj, U, V, params.f, params.loss_f);

function val = yloss(model, split, params)
% YLOSS - Compute the loss on held out test examples from Y

y = split.ytest;
w = repmat(1/length(y), 1, length(y));
V = model.V;
Z = model.Z;
val = evaluate(y, w, split.eyi, split.eyj, V, Z, params.ff, params.loss_ff);

function val = qloss(model, split, params)
% QLOSS - Compute the test loss on held out test examples from Q

q = split.qtest;
w = repmat(1/length(q), 1, length(q));
V = model.V;
R = model.R;
val = evaluate(q, w, split.eqi, split.eqj, V, R, params.fff, params.loss_fff);

function value = evaluate(x, w, ridx, cidx, A, B, f, loss)
% EVALUATE - Computes the weighted loss of predictions given ground truth
%
% This implementation is memory/time efficient. Uses n^2 memory where n is
% the number of test samples. n << numel(AB').
%
% Arguments
%
% x: Value of the ground truth arguments (1 x n)
% w: Weights on the data (1 x n)
% ridx: Row index for each observation in x (1 x n)
% cidx: Column index for each observation in x (1 x n)
% A: Row factors (m x k)
% B: Column factors (r x k)
% f: Prediction link
% loss: Loss to evaluate under, usually matches f.
%
% Return Arguments:
%
% value: The weighted loss of the given observations.

% $$$ value = 0;
% $$$ for i = 1:length(x)
% $$$   r = ridx(i);
% $$$   c = cidx(i);
% $$$   value = value + loss(x(i), w(i), A(r,:)*B(c,:)', f);
% $$$ end
% $$$ value = full(value);

xhat = sum(A(ridx,:).*B(cidx,:),2)';
value = full(loss(x, w, xhat, f));
