function results = data(obj, varargin)
% DATA - All domain specific information for Netflix-IMDB style
% data. This includes extracting data matrices, defining weights, and
% procedures for splitting the data into test and training sets.
%
% Any logic specific to a Netflix-IMDB data set, including weight
% functions, should be maintained here.
%
% The result of this function should be considered immutable. That
% is, changes made outside the scope of this function may or may
% not be accepted.
%
% Arguments:
%
% obj: Instance of a @movie object.
%
% Return Values:
%
% The return value is a structure (results) with the following fields:
%   X: Data matrix for the 3-factor model (m x n)
%   Y: Data matrix for the 3-factor model (n x r)
%   test_train_splits: Function that generates training matrices,
%                      test data, and weight functions that are
%                      zero where the test data was drawn. See
%                      wrapper, below.

% Author: ajit@cs.cmu.edu (Ajit Singh)

[use_actor_as_third_type actor_appearances unused] = ...
    process_options(varargin, 'use_actor', false, 'actor_appearances', 5);
[rating_prediction_ignore_zeros unused] = ...
    process_options(varargin, 'ignore_zeros', false);

fprintf(1, '---------------- Coercing @movie object ---------------------\n');

X = obj.ratings;
if use_actor_as_third_type
  Y = double(logical(obj.actors'));

  threshold = actor_appearances; % needs to vary with the data
  nmovies = rows(Y);
  nactors = cols(Y);
  freq = full(sum(Y,1));
  remove_idx = find(freq < threshold);
  fprintf(1, 'Removing infrequent actors (in < %d movies)\n', threshold);
  Y = Y(:,setdiff(1:nactors, remove_idx));
  fprintf(1, 'Removed %d/%d actors. Kept %d actors\n', length(remove_idx), ...
          nactors, cols(Y));
else
  Y = obj.genres';

  threshold = 0.01;
  ndocs = rows(Y);
  nwords = cols(Y);
  freq = full(sum(Y,1));
  remove_idx = find(freq/ndocs < threshold);
  fprintf(1, 'Removing rare genres (in < %d %% of movies)\n', 100*threshold);
  for i = 1:length(remove_idx)
    fprintf(1, 'Removing infrequent genre: %s\n', ...
            obj.genre_names{remove_idx(i)});
  end
  Y = Y(:,setdiff(1:nwords, remove_idx));
  fprintf(1, 'Removed %d/%d genres\n', length(remove_idx), nwords);
end

% The weight function for the X matrix. The global scaling term
% 1/pex ensures we are measuring per-element loss. The 1/bax term
% balances the total influence of nonzero and zero entries. The
% disx term in [0,1] is a discount factor (ie, disx = 0.1 means
% that zero entries count 1/10 as a much as nonzero entries).

% HACK: The right way to implement the weight functions for X and Y would
% be to use anonymous functions:
%
% wxf = @(x)(1/pex * (1/bax * (x|x) + disx * (~x)));
% wyf = @(y)(1/pey * (1/bay * (y|y) + disy * (~y)));
%
% which could then be passed to weight_function. However, this makes
% evaluting the weights (for example, in glm.m) so slow that it is often
% more expensive than computing the Newton step in practice. A slightly
% faster approach is to use nested functions:
%
% wxf = @nested_weight_x;
%       function value = nested_weight_x(x)
%         value = 1/pex * (1/bax * (x|x) + disx * (~x));
%       end
% wyf = @nested_weight_y;
%       function value = nested_weight_y(y)
%         value = 1/pey * (1/bay * (y|y) + disy * (~y));
%       end
%
% which is slightly faster, but still involves an expensive lookup of
% pex, bax, disx, pey, bay, and disy in its enclosing scope. The best
% solution right now is to declare pex, pey, etc., as global. The suffix
% 'GZPUQ' is a random string attached to minimize the chance of a collison
% in the global namespace.
%
% **** WARNING ****
% The use of global variables means that in a Matlab session you can only
% call data(obj) on one @movie object.

global pex_GZPUQ bax_GZPUQ disx_GZPUQ wxf_GZPUQ;

% pex_GZPUQ: Number of elements of X with weight > 0 (not counting test elt)
% bax_GZPUQ: Relative scaling of zeros vs. non-zeros in matrix X.
% disx_GZPUQ: How much weight to give to zero elts.
% wxf_GZPUQ: Handle to the weight function for X.

pex_GZPUQ = numel(X);
bax_GZPUQ = 1;
if rating_prediction_ignore_zeros
  disx_GZPUQ = 0;
  warning('Zero rating == zero weight');
else
  disx_GZPUQ = density(X);
end
wxf_GZPUQ = @(x)(error('Handle. Do not evaluate'));

global pey_GZPUQ bay_GZPUQ disy_GZPUQ wyf_GZPUQ;
% pey_GZPUQ: Number of elements of Y with weight > 0 (not counting test elt)
% bay_GZPUQ: Relative scaling of zeros vs. non-zeros in matrix Y.
% disy_GZPUQ: How much weight to give to zero elts.
% wyf_GZPUQ: Handle to the weight function for Y.

pey_GZPUQ = numel(Y);
bay_GZPUQ = 1;
disy_GZPUQ = density(Y);
wyf_GZPUQ = @(x)(error('Handle. Do not evaluate'));

results.X = X;
results.Y = Y;
results.test_train_splits = ...
    @(fracx, fracy, nreps)(wrapper(X, Y, fracx, fracy, nreps, wxf_GZPUQ, wyf_GZPUQ));

results.description = sprintf(...
    [ '---------------------- Data Description --------------------\n', ...
      '@movie data set\n',...
      'X: user-movie ratings matrix\n',...
      'Y: movie-genre (binary) matrix\n',...
      'Weight function on X: %s\n',...
      'Weight function on Y: %s\n',...
      '------------------------------------------------------------\n'],...
    sprintf('%s (pex = %g, bax = %g, disx = %g)', ...
            func2str(wxf_GZPUQ), pex_GZPUQ, bax_GZPUQ, disx_GZPUQ), ...
    sprintf('%s (pey = %g, bay = %g, disy = %g)', ...
            func2str(wyf_GZPUQ), pey_GZPUQ, bay_GZPUQ, disy_GZPUQ));

fprintf(1, '-------------------------------------------------------------\n');
end

function results = wrapper(X, Y, fracx, fracy, nreps, wxf, wyf)
% WRAPPER - Split the data matrices X and Y into training and test
% sets.  Generate training matrices, hold out observations for
% testing, and generate a weight function that assigns zero weight to
% the test entries of X and Y.
%
% Implementation Details:
%
% We split the data matrices, hold out observations for testing,
% and generate a weight function for each fold that assigns zero
% weight to the test entries of X and Y.
%
% Arguments
%
% X: Data matrix (m x n)
% Y: Data matrix (n x r)
% fracx: Fraction of nonzero elements of X to hold out.
% fracy: Fraction on nonzero elements of Y to hold out.
% nreps: Number of splits to produce.
% wxf: A weight function that maps an entry of X to a positive value.
% wyf: A weight function that maps an entry of Y to a positive value.
%
% Return Values:
%
% A structure (results) containing the following fields
%   X_train: Training data matrix (m x n)
%   xtest: Vector of held out values of X (1 x num_held_out)
%   exi: i^th entry is the row xtest(i) came from.
%   exj: i^th entry is the column xtest(i) came from.
%   wfcontext: Containing the fields needed to implement standard
%              weights (1/scale ( (x|x) + zero_discount * (~x) ) for
%              both the X and Y matrices.
% and similarly for Y_train, etc ...

global pex_GZPUQ disx_GZPUQ pey_GZPUQ disy_GZPUQ;

assert(fracx > 0);
assert(fracy > 0);
% $$$ assert(nnz(X) < floor(0.5 * numel(X)), ...
% $$$        'We need to assume that less that half of X is nonzero');
% $$$ assert(nnz(Y) < floor(0.5 * numel(Y)), ...
% $$$        'We need to assume that less than half of Y is nonzero');

results = cell(nreps, 1);
for i = 1:nreps
  nhold = ceil(nnz(X) * fracx);
  hold_zeros = ceil(disx_GZPUQ * nhold); % fast sampling wrt data weights
  hold_nnz = nhold - hold_zeros;

  [res.X_train res.xtest res.exi res.exj res.wxf Exclude Exclude_t] = ...
      split_matrix(X, hold_zeros, hold_nnz, @weight_function, wxf);
  wx_context = [ ];
  wx_context.pex = pex_GZPUQ;
  wx_context.disx = disx_GZPUQ;
  wx_context.Exclude = Exclude;
  wx_context.Exclude_trans = Exclude_t;
  res.wx_context = wx_context;
  fprintf(1, 'Holding out %d X-links\n', nhold);

  nhold = ceil(nnz(Y) * fracy);
  hold_zeros = ceil(disy_GZPUQ * nhold);
  hold_nnz = nhold - hold_zeros;
  [res.Y_train res.ytest res.eyi res.eyj res.wyf Exclude Exclude_t] = ...
      split_matrix(Y, hold_zeros, hold_nnz, @weight_function, wyf);
  wy_context = [ ];
  wy_context.pey = pey_GZPUQ;
  wy_context.disy = disy_GZPUQ;
  wy_context.Exclude = Exclude;
  wy_context.Exclude_trans = Exclude_t;
  res.wy_context = wy_context;
  fprintf(1, 'Holding out %d Y-links\n', nhold);

  res.eval_x = @mae;
  res.eval_y = @mae;
  results{i} = res;
  fprintf(1, 'Test error func: X = %s, Y = %s\n', ...
          func2str(res.eval_x), func2str(res.eval_y));
end
end

function weight = weight_function(x, i, j, Exclude, Exclude_trans, f)
% WEIGHT_FUNCTION - Evaluate a weight function to f(x,i,j) if (i,j)
% are not in the exclusion list. Otherwise evaluate the weight to
% zero.
%
% Implementation Details:
%
% This function is meant to model a weight function that evaluates
% to zero where the entry was held out (or is otherwise
% missing). This function will be called on every update in glm.m,
% and so efficiency is important.
%
% Arguments:
%
% x: Vector of data. (1 x m)
% Exclude: Binary matrix where 1 mean the position is held-out.
% Exclude_trans: The transpose of Exclude. We pass it separately because
%                accessing rows of a sparse matrix is often 100x slower than
%                accessing a column in the transpose.
% [i j]: Row and columns of x in the data matrix. (1 x m, 1 x m)
% [exclude_i exclude_j]: Entries of the weight matrix that must be zero
% f: Strictly positive weight function
%
% Return Values:
%
% weight: Weight vector (1 x m) that is zero only where (i,j) was
%         in the exclude list.

global wxf_GZPUQ pex_GZPUQ bax_GZPUQ disx_GZPUQ;
global wyf_GZPUQ pey_GZPUQ bay_GZPUQ disy_GZPUQ;

weight = NaN * ones(size(x));
if isequal(f, wxf_GZPUQ)
  %weight = 1/pex_GZPUQ * (1/bax_GZPUQ * (x|x) + disx_GZPUQ * (~x));
  idx = (x ~= 0);
  weight(idx) = 1/(pex_GZPUQ * bax_GZPUQ);
  weight(~idx) = disx_GZPUQ / pex_GZPUQ;
elseif isequal(f, wyf_GZPUQ)
  y = x;
  %weight = 1/pey_GZPUQ * (1/bay_GZPUQ * (y|y) + disy_GZPUQ * (~y));
  idx = (y ~= 0);
  weight(idx) = 1/(pey_GZPUQ * bay_GZPUQ);
  weight(~idx) = disy_GZPUQ / pey_GZPUQ;
else
  error('Something is wrong with wrapper, in all likelihood.');
end

if not(isempty(i)) && not(isempty(j))
  assert(length(i) == length(x));
  assert(length(j) == length(x));

  % Zero the weight at positions held out for testing. To avoid expensive
  % calls to sub2ind, we check whether the data is a row or column of a
  % data matrix and specialize accordingly.
  if all(i == i(1))
    weight(logical(Exclude_trans(:,i(1)))) = 0;
  elseif all(j == j(1))
    weight(logical(Exclude(:,j(1)))) = 0;
  else
    exclude = logical(Exclude(sub2ind(size(Exclude), i, j)));
    weight(exclude) = 0;
  end
end
weight = full(weight);

end





