function results = data4f(obj, varargin)
% DATA4F - All domain specific information for Netflix-IMDB style
% data. This includes extracting data matrices, defining weights, and
% procedures for splitting the data into test and training sets.
%
% Any logic specific to a Netflix-IMDB data set, including weight
% functions, should be maintained here.
%
% The result of this function should be considered immutable. That
% is, changes made outside the scope of this function may or may
% not be accepted.
%
% This object is compatable with @movie/data.m. That is, you can use this
% object whereever a 3-factor data set is required (The data and splits for
% glm4f can also be used in glm).
%
% Arguments:
%
% obj: Instance of a @movie object.
%
% Return Values:
%
% The return value is a structure (results) with the following fields:
%   X: Data matrix for the 4-factor model (m x n)
%   Y: Data matrix for the 4-factor model (n x r)
%   test_train_splits: Function that generates training matrices,
%                      test data, and weight functions that are
%                      zero where the test data was drawn. See
%                      wrapper, below.

% Author: ajit@cs.cmu.edu (Ajit Singh)

X = obj.ratings;
Y = obj.genres';
% TODO(ajit): Why are some entries in Q not zero or one ? (eg, in movies_500
% some of the entries are 2 or 3).
Q = obj.actors';
Q = double(logical(Q));

[actor_appearances unused] = process_options(varargin, 'actor_appearances',...
                                             5);
[rating_prediction_ignore_zeros unused] = ...
    process_options(varargin, 'ignore_zeros', false);

fprintf(1, '---------------- Coercing @movie object ---------------------\n');

threshold = 0.01;
ndocs = rows(Y);
nwords = cols(Y);
freq = full(sum(Y,1));
remove_idx = find(freq/ndocs < threshold);
fprintf(1, 'Removing rare genres (in < %d %% of movies)\n', 100*threshold);
for i = 1:length(remove_idx)
  fprintf(1, 'Removing infrequent genre: %s\n', ...
          obj.genre_names{remove_idx(i)});
end
Y = Y(:,setdiff(1:nwords, remove_idx));
fprintf(1, 'Removed %d/%d genres\n', length(remove_idx), nwords);

threshold = actor_appearances; % needs to vary with the data
nmovies = rows(Q);
nactors = cols(Q);
freq = full(sum(Q,1));
remove_idx = find(freq < threshold);
fprintf(1, 'Removing infrequent actors (in < %d movies)\n', threshold);
Q = Q(:,setdiff(1:nactors, remove_idx));
fprintf(1, 'Removed %d/%d actors. Kept %d actors\n', length(remove_idx), ...
        nactors, cols(Q));

global pex_TYUP disx_TYUP wxf_TYUP;

pex_TYUP = numel(X);
if rating_prediction_ignore_zeros
  disx_TYUP = 0;
else
  disx_TYUP = density(X);
end
wxf_TYUP = @(x)(error('Handle. Do not evaluate'));

global pey_TYUP disy_TYUP wyf_TYUP;

pey_TYUP = numel(Y);
disy_TYUP = density(Y);
wyf_TYUP = @(y)('Handle. Do not evaluate');

global peq_TYUP disq_TYUP wqf_TYUP;

peq_TYUP = numel(Q);
disq_TYUP = density(Q);
wqf_TYUP = @(q)('Handle. Do not evaluate');

results.X = X;
results.Y = Y;
results.Q = Q;
results.test_train_splits = ...
    @(fx, fy, fq, nreps)(wrapper(X, Y, Q, fx, fy, fq, nreps, wxf_TYUP, wyf_TYUP, wqf_TYUP));

results.description = sprintf(...
    [ '---------------------- Data Description --------------------\n', ...
      '@movie data set\n',...
      'X: user-movie ratings matrix\n',...
      'Y: movie-genre (binary) matrix\n',...
      'Weight function on X: %s\n',...
      'Weight function on Y: %s\n',...
      'Weight function on Q: %s\n',...
      '------------------------------------------------------------\n'],...
    sprintf('%s (pex = %g, disx = %g)', ...
            func2str(wxf_TYUP), pex_TYUP, disx_TYUP),...
    sprintf('%s (pey = %g, disy = %g)', ...
            func2str(wyf_TYUP), pey_TYUP, disy_TYUP), ...
    sprintf('%s (peq = %g, disq = %g)', ...
            func2str(wqf_TYUP), peq_TYUP, disq_TYUP));

fprintf(1, '-------------------------------------------------------------\n');

function results = wrapper(X, Y, Q, fracx, fracy, fracq, nreps, ...
                           wxf, wyf, wqf, varargin)
% WRAPPER - Split the data matrices X and Y into training and test
% sets.  Generate training matrices, hold out observations for
% testing, and generate a weight function that assigns zero weight to
% the test entries of X and Y.
%
% Implementation Details:
%
% We split the data matrices, hold out observations for testing,
% and generate a weight function for each fold that assigns zero
% weight to the test entries of X and Y.
%
% Arguments
%
% X: Data matrix (m x n)
% Y: Data matrix (n x r)
% fracx: Fraction of nonzero elements of X to hold out.
% fracy: Fraction on nonzero elements of Y to hold out.
% nreps: Number of splits to produce.
% wxf: A weight function that maps an entry of X to a positive value.
% wyf: A weight function that maps an entry of Y to a positive value.
%
% Return Values:
%
% A structure (results) containing the following fields
%   X_train: Training data matrix (m x n)
%   xtest: Vector of held out values of X (1 x num_held_out)
%   exi: i^th entry is the row xtest(i) came from.
%   exj: i^th entry is the column xtest(i) came from.
% and similarly for Y_train, etc ...

global pex_TYUP disx_TYUP;
global pey_TYUP disy_TYUP;
global peq_TYUP disq_TYUP;

assert(fracx > 0);
assert(fracy > 0);
assert(fracq > 0);

results = cell(nreps, 1);
for i = 1:nreps
  nhold = ceil(nnz(X) * fracx);
  hold_zeros = ceil(disx_TYUP * nhold);
  hold_nnz = nhold - hold_zeros;
  [res.X_train res.xtest res.exi res.exj res.wxf Exclude Exclude_t] = ...
      split_matrix(X, hold_zeros, hold_nnz, @weight_function, wxf);
  wx_context = [ ];
  wx_context.pex = pex_TYUP;
  wx_context.disx = disx_TYUP;
  wx_context.Exclude = Exclude;
  wx_context.Exclude_trans = Exclude_t;
  res.wx_context = wx_context;
  fprintf(1, 'Holding out %d X-links\n', nhold);

  nhold = ceil(nnz(Y) * fracy);
  hold_zeros = ceil(disy_TYUP * nhold);
  hold_nnz = nhold - hold_zeros;
  [res.Y_train res.ytest res.eyi res.eyj res.wyf Exclude Exclude_t] = ...
      split_matrix(Y, hold_zeros, hold_nnz, @weight_function, wyf);
  wy_context = [ ];
  wy_context.pey = pey_TYUP;
  wy_context.disy = disy_TYUP;
  wy_context.Exclude = Exclude;
  wy_context.Exclude_trans = Exclude_t;
  res.wy_context = wy_context;
  fprintf(1, 'Holding out %d Y-links\n', nhold);

  nhold = ceil(nnz(Q) * fracq);
  hold_zeros = ceil(disq_TYUP * nhold);
  hold_nnz = nhold - hold_zeros;
  [res.Q_train res.qtest res.eqi res.eqj res.wqf Exclude Exclude_t] = ...
      split_matrix(Q, hold_zeros, hold_nnz, @weight_function, wqf);
  wq_context = [ ];
  wq_context.peq = peq_TYUP;
  wq_context.disq = disq_TYUP;
  wq_context.Exclude = Exclude;
  wq_context.Exclude_trans = Exclude_t;
  res.wq_context = wq_context;
  fprintf(1, 'Holding out %d Q-links\n', nhold);

  res.eval_x = @mae;
  res.eval_y = @mae;
  res.eval_q = @mae;
  results{i} = res;
  fprintf(1, 'Test error func: X = %s, Y = %s, Q = %s\n', ...
          func2str(res.eval_x), func2str(res.eval_y), func2str(res.eval_q));
end

function weight = weight_function(x, i, j, Exclude, Exclude_trans, f)
% WEIGHT_FUNCTION - Evaluate a weight function to f(x,i,j) if (i,j)
% are not in the exclusion list. Otherwise evaluate the weight to
% zero.
%
% Implementation Details:
%
% This function is meant to model a weight function that evaluates
% to zero where the entry was held out (or is otherwise
% missing). This function will be called on every update in glm.m,
% and so efficiency is important.
%
% Arguments:
%
% x: Vector of data. (1 x m)
% [i j]: Row and columns of x in the data matrix. (1 x m, 1 x m)
% [exclude_i exclude_j]: Entries of the weight matrix that must be zero
% f: Strictly positive weight function
%
% Return Values:
%
% weight: Weight vector (1 x m) that is zero only where (i,j) was
%         in the exclude list.

global pex_TYUP disx_TYUP wxf_TYUP;
global pey_TYUP disy_TYUP wyf_TYUP;
global peq_TYUP disq_TYUP wqf_TYUP;

weight = NaN * ones(size(x));
if isequal(f, wxf_TYUP)
  idx = (x ~= 0);
  weight(idx) = 1/(pex_TYUP);
  weight(~idx) = disx_TYUP / pex_TYUP;
elseif isequal(f, wyf_TYUP)
  idx = (x ~= 0);
  weight(idx) = 1/pey_TYUP;
  weight(~idx) = disy_TYUP / pey_TYUP;
elseif isequal(f, wqf_TYUP)
  idx = (x ~= 0);
  weight(idx) = 1/peq_TYUP;
  weight(~idx) = disq_TYUP / peq_TYUP;
else
  error('Something is wrong with wrapper, in all likelihood');
end

if not(isempty(i)) && not(isempty(j))
  assert(length(i) == length(x));
  assert(length(j) == length(x));

  if all(i == i(1))
    weight(logical(Exclude_trans(:,i(1)))) = 0;
  elseif all(j == j(1))
    weight(logical(Exclude(:,j(1)))) = 0;
  else
    exclude = logical(Exclude(sub2ind(size(Exclude), i, j)));
    weight(exclude) = 0;
  end
end
weight = full(weight);
