function obj = movie(varargin)
% MOVIE - Constructor for a @movie object.
%
% Usage:
%
% 1. As a default constructor (see init_fields, below)
%
% obj = movie
%
% 2. As a copy constructor
%
% obj = movie;
% copy_obj = movie(obj);
%
% 3. Load a movie object from text files
%
% obj = movie('~/imdb/');

% Author: ajit@cs.cmu.edu (Ajit Singh)

classname = 'movie';
if nargin == 0
  obj = init_fields;
  obj = class(obj, classname);
  return;
end
firstArg = varargin{1};
if isa(firstArg, classname)
  obj = firstArg;
  return;
end

obj = init_fields;
obj = class(obj,classname);

if nargin == 1
  % TODO(ajit): Use fullfile instead.
  dirname = varargin{1};
  if dirname(end) ~= '/'
    dirname = strcat(dirname,'/');
  end
  [ig,obj.genre_names] = textread(strcat(dirname,'genres.txt'),'%d %s');
  [ig,obj.movie_names] = textread(strcat(dirname,'movie_names.txt'),'%d %s');
  [ig,obj.talent_names] = textread(strcat(dirname,'people_names.txt'),'%d %s');
  [ig,obj.netflix_id] = textread(strcat(dirname,'user_custid.txt'),'%d %s');
  obj.Nu = length(obj.netflix_id);
  obj.Ng = length(obj.genre_names);
  obj.Nm = length(obj.movie_names);
  obj.Nt = length(obj.talent_names);

  obj.genres = load_sparse_matrix(strcat(dirname,'movie_genres.txt'));
  obj.genres = obj.genres';
  [n,m] = size(obj.genres);
  if (n ~= obj.Ng)
    error('Loaded %d, expected %d genres\n',n,obj.Ng);
  end
  if (m ~= obj.Nm)
    error('Loaded %d, expected %d movies\n',m,obj.Nm);
  end

  obj.actors = load_sparse_matrix(strcat(dirname,'actor_movies.txt'));
  [n,m] = size(obj.actors);
  if (n ~= obj.Nt)
    error('Loaded %d, expected %d people\n',n,obj.Nt);
  end
  if (m ~= obj.Nm)
    error('Loaded %d, expected %d movies\n',m,obj.Nm);
  end

  obj.directors = load_sparse_matrix(strcat(dirname,'director_movies.txt'));
  [n,m] = size(obj.directors);
  if (n ~= obj.Nt)
    error('Loaded %d, expected %d people\n',n,obj.Nt);
  end
  if (m ~= obj.Nm)
    error('Loaded %d, expected %d movies\n',m,obj.Nm);
  end

  obj.movies_follows = load_sparse_matrix(strcat(dirname,'movie_movies_follows.txt'));
  [n,m] = size(obj.movies_follows);
  if (n ~= obj.Nm)
    error('Loaded %d, expected %d movies\n',n,obj.Nm);
  end
  if (m ~= obj.Nm)
    error('Loaded %d, expected %d movies\n',m,obj.Nm);
  end

  obj.movies_remake = load_sparse_matrix(strcat(dirname,'movie_movies_remake.txt'));
  [n,m] = size(obj.movies_remake);
  if (n ~= obj.Nm)
    error('Loaded %d, expected %d movies\n',n,obj.Nm);
  end
  if (m ~= obj.Nm)
    error('Loaded %d, expected %d movies\n',m,obj.Nm);
  end

  obj.movies_spinoff = load_sparse_matrix(strcat(dirname,'movie_movies_spinoff.txt'));
  [n,m] = size(obj.movies_spinoff);
  if (n ~= obj.Nm)
    error('Loaded %d, expected %d movies\n',n,obj.Nm);
  end
  if (m ~= obj.Nm)
    error('Loaded %d, expected %d movies\n',m,obj.Nm);
  end

  obj.ratings = load_sparse_matrix(strcat(dirname,'user_movies.txt'));
  [n,m] = size(obj.ratings);
  if (n ~= obj.Nu)
    error('Loaded %d, expected %d movies\n',n,obj.Nm);
  end
  if (m ~= obj.Nm)
    error('Loaded %d, expected %d movies\n',m,obj.Nm);
  end

  fprintf('Number of users = %d\n',obj.Nu);
  fprintf('Number of genres = %d\n',obj.Ng);
  fprintf('Number of movies = %d\n',obj.Nm);
  fprintf('Number of actors = %d\n',obj.Nt);

  % Remove any movies with no ratings
  movie_idx = find(sum(obj.ratings,1));
  if ( obj.Nm - length(movie_idx) > 0 )
    fprintf('Removed %d movies with no ratings\n',obj.Nm - length(movie_idx));
  end
  obj.ratings = obj.ratings(:,movie_idx);
  obj.genres = obj.genres(:,movie_idx);
  obj.actors = obj.actors(:,movie_idx);
  obj.directors = obj.directors(:,movie_idx);
  obj.movies_follows = obj.movies_follows(movie_idx,movie_idx);
  obj.movies_remake = obj.movies_remake(movie_idx,movie_idx);
  obj.movies_spinoff = obj.movies_spinoff(movie_idx,movie_idx);
  obj.movie_names = obj.movie_names(movie_idx);
  obj.Nm = length(movie_idx);

  % Remove any users with no ratings
  user_idx = find(sum(obj.ratings,2));
  if ( obj.Nu - length(user_idx) > 0 )
    fprintf('Removed %d users with no ratings\n',obj.Nu - length(user_idx));
  end
  obj.ratings = obj.ratings(user_idx,:);
  obj.netflix_id = obj.netflix_id(user_idx);
  obj.Nu = length(user_idx);

  % Remove any actors who are neither actors nor directors
  talent_idx = find(sum(obj.actors + obj.directors,2));
  if (obj.Nt - length(talent_idx) > 0)
    fprintf('Removed %d people who are neither actors nor directors\n',...
            obj.Nt - length(talent_idx));
  end
  obj.actors = obj.actors(talent_idx,:);
  obj.directors = obj.directors(talent_idx,:);
  obj.talent_names = obj.talent_names(talent_idx);
  obj.Nt = length(talent_idx);
end

function obj = init_fields()
% INIT_FIELDS - Default initializer for a @movie object.

obj.Nu = 0; % number of netflix customers (users)
obj.Ng = 0; % number of genres (28)
obj.Nm = 0; % number of movies
obj.Nt = 0; % number of performers (talent)
obj.ratings = [];  % users x movies
obj.genres = [];   % genres x movies
obj.actors = [];   % talent x movies
obj.directors = []; % talent x movies
obj.movies_follows = [];   % movies x movies
obj.movies_remake = []; %movies x movies
obj.movies_spinoff = []; % movies x movies
obj.netflix_id = [];       % array(i) is the netflix customer id for user i
obj.movie_names = [];    % array{i} is the name
obj.genre_names = []; % array{i} is the name of the genre
obj.talent_names = [];  % array{i} is the name of an actor,director

