function newobj = shrink(obj, nusers, nmovies, ...
                         user_minratings, user_maxratings, ...
                         movie_minratings, movie_maxratings, ...
                         must_have_genre, must_have_actors)
% SHRINK - Create a smaller version of a @movie object by sampling
% users.
%
% Arguments:
%
% obj: An instance of class @movie.
% nusers: Number of users to choose [0..Inf] (Inf means all users)
% nmovies: Number of movies to choose [0..Inf] (Inf means all movies)
% user_minratings: Keep only users who rated at least this many movies.
% movie_minratings: Keep only movies with at least this many ratings.
% must_have_genre: If true, select only movies that have at least
% one genre. If false, some of the movies may have no genres.
%
% Return Values:
%
% newobj: An instance of class @movie with the given number of
% users. Movies unseen by any of the chosen users have been removed,
% as well as any genres that are now empty.

% Author: ajit@cs.cmu.edu (Ajit Singh)

error(nargchk(9, 9, nargin));

% $$$ error(nargchk(3, 6, nargin));
% $$$ if nargin < 6
% $$$   must_have_genre = true;
% $$$ end
% $$$ if nargin < 5
% $$$   movie_minratings = 0;
% $$$ end
% $$$ if nargin < 4
% $$$   user_minratings = 0;
% $$$ end

assert(nusers > 0, 'nusers must be > 0');
assert(nmovies > 0, 'nmovies must be > 0');
assert(user_minratings >= 0, 'user_minratings must be > 0');
assert(movie_minratings >= 0, 'movie_minratings must be > 0');
assert(movie_minratings <= movie_maxratings);
assert(user_minratings <= user_maxratings);

fprintf(1, 'Selecting %d users, %d movies\n', nusers, nmovies);
fprintf(1, 'All users must have [%d, %d] ratings\n', ...
        user_minratings, user_maxratings);
fprintf(1, 'All movies must be rated by at least [%d, %d] selected users\n', ...
        movie_minratings, movie_maxratings);
if must_have_genre
  fprintf(1, 'Selecting only movies with at least one genre\n');
else
  fprintf(1, 'Some movies may have no genre info\n');
end
if must_have_actors
  fprintf(1, 'Selecting only movies with at least one actor\n');
else
  fprintf(1, 'Some movies may have no actor info\n');
end

newobj = movie(obj);

users = find_users_with_ratings(obj, user_minratings, user_maxratings);
if length(users) < nusers && isfinite(nusers)
  error('RFA:Boundary', ...
        'Not enough users with [%d, %d] ratings (want %d, are %d)',...
        user_minratings, user_maxratings, nusers, length(users));
end
if isfinite(nusers)
  users = sample(users, nusers);
end

nratings_movie = sum(obj.ratings(users, :));
movie_condition = (nratings_movie >= movie_minratings) & ...
                  (nratings_movie <= movie_maxratings);
if must_have_genre
  movie_condition = movie_condition & (sum(obj.genres, 1) > 0);
end
if must_have_actors
  movie_condition = movie_condition & (sum(obj.actors, 1) > 0);
end
movies = find(movie_condition);

if length(movies) < nmovies && isfinite(nmovies)
  error('RFA:Boundary', ...
        'Not enough movies matching conditions (want %d, are %d)', ...
        nmovies, length(movies));
end
if isfinite(nmovies)
  movies = sample(movies, nmovies);
end

% NOTE: rows of actors and directors can be empty (eg, if a movie
% has a director listed, but not actors listed, or vice versa).
newobj.ratings = obj.ratings(users, movies);
[newobj.genres genre_ids] = remove_empty_rows(obj.genres(:, movies));
[ignore actor_ids] = remove_empty_rows(obj.actors(:, movies));
[ignore director_ids] = remove_empty_rows(obj.directors(:, movies));
talent_ids = union(actor_ids, director_ids);
newobj.actors = obj.actors(talent_ids, movies);
newobj.directors = obj.directors(talent_ids, movies);

% TODO(ajit): Consider the following issue:
% Only preserve follows links between the selected movies. Since we
% are sampling randomly chances are that this matrix is
% empty. Maybe not a huge issue since there are only ~3000 sequel links.
newobj.movies_follows = obj.movies_follows(movies, movies);
newobj.movies_remake = obj.movies_remake(movies, movies);
newobj.movies_spinoff = obj.movies_spinoff(movies, movies);

newobj.netflix_id = obj.netflix_id(movies);
newobj.movie_names = obj.movie_names(movies);
newobj.genre_names = obj.genre_names(genre_ids);
newobj.talent_names = obj.talent_names(talent_ids);

newobj.Nu = rows(newobj.ratings);
newobj.Ng = rows(newobj.genres);
newobj.Nm = cols(newobj.ratings);
newobj.Nt = rows(newobj.actors);

function [M keep_idx] = remove_empty_rows(M)

keep_idx = find(sum(M, 2));
M = M(keep_idx, :);

function [M keep_idx] = remove_empty_cols(M);

keep_idx = find(sum(M, 1));
M = M(:, keep_idx);
