function newobj = shrink(obj, papers, mode)
% SHRINK - Create a new @nips object containing only the given papers.
%
% We guarantee that each author is on at least one paper, and that
% each word belongs to each paper.
%
% Arguments:
%
% obj: An instance of a @nips class.
% papers: A set of paper indices, ie a subset of [1..obj.Np].
% mode: If equal to 'sample', then treat papers as the number of papers
%       to sample from obj.
%
% Return Values:
%
% newobj: An version of obj where the given papers were removed,
% along with authors who belong to none of the remaining papers,
% and words that occur in none of the remaining papers.
%
% Example:
%
%   papers = sample(1:obj.Np, ceil(obj.Np/2));
%   newobj = shrink(obj, papers);
%   ... or equivalently,
%   newobj = shrink(obj, ceil(obj.Np/2), 'sample');

% Author: ajit@cs.cmu.edu (Ajit Singh)

error(nargchk(2, 3, nargin));
if nargin == 2
  mode = [ ];
end

if strcmpi(mode, 'sample')
  assert(papers > 0, 'No point in sampling zero papers');
  assert(papers <= obj.Np, 'Not enough papers (want %d, have %d)', ...
         papers, obj.Np);
  papers = sample(1:obj.Np, papers);
elseif length(papers) == 1
  warning('RFA:Sanity', 'You are sampling one paper. See the mode argument');
end

newobj = obj;
newobj.Np = length(papers);
newobj.adocs = obj.adocs(:, papers);
newobj.counts = obj.counts(:, papers);
newobj.docnames = obj.docnames(papers);
newobj.panum = obj.panum(papers);
newobj.plengths = obj.plengths(papers);

% Remove all the authors who have no documents
authors = find(sum(newobj.adocs, 2));
newobj.Na = length(authors);
newobj.adocs = newobj.adocs(authors, :);
newobj.anames = newobj.anames(authors);
newobj.apapers = newobj.apapers(authors);

% Remove all the words that are not in any document
documents = find(sum(newobj.counts, 2));
newobj.counts = newobj.counts(documents, :);
newobj.wc = sum(newobj.counts, 2);
newobj.wl = newobj.wl(documents);
newobj.Nw = size(newobj.counts, 1);

if any((sum(newobj.adocs, 1) == 0) & (sum(newobj.counts, 1) == 0))
  warning('RFA:Postcondition', ...
          'Some documents have neither author nor word information');
end
