%EVAL_RSCORE Compute the R-score (expected rank utility).
%   EVAL_RSCORE(obj,entities,data,utility,alpha) computes the
%   R-score (expected rank utility) [Breese 1998]. This function takes
%       obj - ranking object
%       entities - the true (held-out) links. A subset of obj.rank
%       data - the data set of interest
%       utility - utility function (that uses the data object)
%       alpha - mean number of recommendations taken

function rscore = eval_rscore(obj,entities,data,utility,alpha)

check_size_invariant(obj);
check_score_invariant(obj);
check_uniqueness_invariant(obj);
check_entity_invariant(obj,entities);
if isempty(entities)
    error('@ranking/score: no entities, cannot compute average rank.');
end
if alpha < 1 || int32(alpha) ~= alpha
    error('@ranking/score: alpha = %f must be integral and > 1',alpha);
end

% Ranking must be sorted
obj = sort(obj);

% Compute the max_rscore
u = utility(data,entities);
sorted_u = sort(u,'descend');
weight_u = power(1/2,(0:length(sorted_u)-1)/alpha);
max_rscore = dot(sorted_u,weight_u);

% Compute the rscore
u = utility(data,obj.rank);
act_rscore = compute_rscore(obj.rank,entities,u,alpha);
rscore = act_rscore/max_rscore;
if rscore > 1.0 || rscore < 0.0
    error('rscore not in [0,1]')
end

% ranking[1...N]
% entities are the true links
% utility are the utilities
function r = compute_rscore(ranking,entities,utility,alpha)

r = 0.0;
for e = entities
    j = find(ranking==e);
    r = r + power(1/2,(j-1)/alpha) * utility(j);
end
