function [model train_time] = ch_em(N,A,params,varargin)
% CH_EM - Learn a pLSI-pHITS model using tempered EM
%
% [model train_time] = ch_em(N, A, params, varargin)
% Implements the Cohn and Hofmann (NIPS-01) paper for a joint document-citation
% model which extends pLSI to the two matrix case. The matrices correspond to
% words x documents and authors x documents
%
% Arguments:
%
% N: Word counts for each document (words x training_papers)
% A: Authorship records for each document (authors x training_papers)
% params: Model and optimization parameters. Takes the same @glmparams
%         objects that glm.m takes and uses only the following fields:
%         params.dims - embedding dimension, K.
%         params.alpha - mixing weight, alpha.
%         params.maxiter - maximum number of EM iterations
%          (in each call to impl_em, minBeta and decayrate govern the
%           maximum number of tempering steps.)
% varargin: Optional arguments
%           minBeta: minimum tempering value in [0,1] (default: 0.8)
%           initBeta: starting tempering value in [0,1] (default: 1.0)
%           decayrate: fraction by which beta is reduced at each tempering
%                      step (default: 0.95)
%           tf, af: Fixed values for TermFactors and AuthFactors. If both
%                   are defined then we only update document mixing factors,
%                   which is useful for folding in new documents.
%           temper_frac: Fraction of non-zero entries of N and A to hold
%                        out for tempering (default: 0.1)
%
% Return Values:
%
% model: The final trained model, a structure containing three fields:
%        TermFactors, AuthFactor, and DocMixing which correspond to
%        the p(term|latent), p(author|latent), and p(latent|doc)
%        respectively. The dimensions are
%        K x terms: Term Factors
%        K x authors: Author Factors
%        K x training_docs: Doc Mixing Factors
%        scalar: alpha, weight in favour of N
%        scalar: beta
% train_time: The total CPU time required to train the model. Includes both
%             the tempered and non-tempered phases of EM.

% Author: Ajit Singh (ajit@cs.cmu.edu)

% TODO(ajit): Does normalizing, and then splitting the tempering set help ?
% TODO(ajit): Add support for params.debug to suppress some messages.

error(nargchk(3, Inf, nargin));

K = params.dims;
alpha = params.alpha;
[minBeta unused] = process_options(varargin, 'minBeta', 0.8);
[initBeta unused] = process_options(varargin, 'initBeta', 1.0);
[tf af unused] = process_options(varargin, 'tf', [ ], 'af', [ ]);
[temper_frac unused] = process_options(varargin, 'temper_frac', 0.10);
[decayrate unused] = process_options(varargin, 'decayrate', 0.95);
[maxiter unused] = process_options(varargin, 'maxiter', params.maxiter);
[maxiter_folding unused] = process_options(varargin, 'maxiter_folding', ...
                                           params.maxiter);
[threshold unused] = process_options(varargin, 'threshold', 10e-2);

assert(isbetween(alpha, 0, 1, 'closed'), ...
       'alpha = %g. Expected to be in [0,1]', alpha);

assert(isbetween(minBeta, 0, 1, 'closed'), ...
       'minBeta = %g. Expected to be in [0,1]', minBeta);

assert(isbetween(initBeta, 0, 1, 'closed'), ...
       'minBeta = %g. Expected to be in [0,1]', initBeta);

assert(isbetween(temper_frac, 0, 1, 'closed'), ...
       'temper_frac = %g. Expected to be in [0,1]', temper_frac);

assert(not(xor(isempty(af),isempty(tf))), ...
       'Cannot define only one of af and tf');

init_time = cputime;

% Compute dimensions with respect to the test set.
Nw = rows(N);
Np = cols(N);
Na = rows(A);

% Hold out a fraction of the links in N and A for use with tempered EM.
[N, Nt] = select_tempering_data(N,temper_frac);
[A, At] = select_tempering_data(A,temper_frac);
%debug_entities_with_no_information(N, A);

% Renormalize each column so that N(i,j) is the proportion of term ti in
% document dj. Similarly renormalize each column so that A(i,j) is
% the proportion of authors ai in document dj (ie, fractional authorship)
N = safe_normalize_cols(N);
A = safe_normalize_cols(A);
Nt = safe_normalize_cols(Nt);
At = safe_normalize_cols(At);

% Initialize model parameters
model.alpha = alpha;
model.beta = initBeta;
model.TermFactors = normalize_rows(rand(K,Nw)); % p(ti|zk)
model.AuthFactors = normalize_rows(rand(K,Na)); % p(al|zk)
model.DocMixing = normalize_cols(rand(K,Np)); % p(zk|dj)
[init_model unused] = process_options(varargin, 'init_model', [ ]);
if not(isempty(init_model))
  model.TermFactors = init_model.TermFactors;
  model.AuthFactors = init_model.AuthFactors;
  model.DocMixing = init_model.DocMixing;
end

% Initialize optimization parameters. If one wants to fold in new documents,
% then tf and af have to be specified.
lparams.maxiter = maxiter;
lparams.threshold = threshold;
lparams.earlystop = true;
lparams.fold = false;
if any(any(tf)) || any(any(af))
  WarnIf(model.beta == 1.0, ...
         'On folding iterations model.beta == initBeta, no tempering.');
  lparams.fold = true; % no TEM steps if we are folding in test samples
  lparams.maxiter = maxiter_folding;
  model.TermFactors = tf;
  model.AuthFactors = af;
end
init_model = model;

if not(lparams.fold)
    fprintf(1,'------ pLSI-pHITS Information --------------\n');
    fprintf(1,'Training set (Nw = %d, Np = %d, Na = %d)\n',Nw,Np,Na);
    fprintf(1,'Number of author-document links: %d\n',nnz(A));
    fprintf(1,'Number of document-word links: %d\n',nnz(N));
    fprintf(1,'Number of author-document links for tempering: %d\n',nnz(At));
    fprintf(1,'Number of document-word links for tempering: %d\n',nnz(Nt));
    fprintf(1,'Weighting in favor of words over authors (alpha) = %.2f\n', ...
            alpha);
    fprintf(1,'Lower bound on entropic regularization (minBeta) = %.2f\n', ...
            minBeta);
    fprintf(1,'Number of latent factors (K) = %d\n',K);
    fprintf(1,'Decay rate (eta) = %.2f\n',decayrate);
    fprintf(1,'-------------------------------------------\n');
end

% Run Tempered EM steps as long as beta is above minBeta and we do not
% overfit between EM iterations (holding out tempering links allows us to
% test for overfitting). If we are folding, do not use tempered EM.
totalsteps = 0;
if not(lparams.fold)
  iter = 1;
  threshold = 0.5;
  gain = -Inf;
  while gain < -threshold && model.beta > minBeta
    fprintf(1,'Tempering EM with beta = %.3f\n',model.beta);
    [model,steps,gain,ll] = impl_em(N,A,Nt,At,K,model,lparams,varargin{:});
    fprintf('Tempering step info (iter %2d): steps = %d, gain = %.5f\n\n', ...
            iter,steps,gain);
    if gain < 0
      model.beta = decayrate * model.beta;
    end
    totalsteps = totalsteps + steps;
  end
end

% Run some final iterations of EM on all the data
% NOTE: We reset the model so that maxiter has the same semantics between
% glm* and ch_em (ie, don't save the model found during tempering, just use
% tempering to select beta).
fprintf(1,'Final EM steps\n');
%--
beta = model.beta;
model = init_model;
model.beta = beta;
%--
lparams.earlystop = false;
[model,steps,gain,ll] = impl_em(N+Nt,A+At,Nt,At,K,model,lparams,varargin{:});
totalsteps = totalsteps + steps;

fprintf(1, 'Overall number of EM steps: %d\n', totalsteps);
train_time = cputime - init_time;

function [Xtrain,Xtemper] = select_tempering_data(X,frac)
% Hold out a fraction of the nonzero entries of X in Xtemper. A postcondition
% is that X == Xtrain + Xtemper.

[Xi,Xj,Xv] = find(X); Xlen = nnz(X);
[temperIdx,trainIdx] = random_split(1:Xlen,max(1,round(frac*Xlen)));
Xtrain = sparse(Xi(trainIdx),Xj(trainIdx),Xv(trainIdx),size(X,1),size(X,2));
Xtemper = sparse(Xi(temperIdx),Xj(temperIdx),Xv(temperIdx),size(X,1),size(X,2));

function [newmodel,iter,gain,ll] = impl_em(N,A,Nt,At,K,model,lparams,varargin)
% Implements the EM update.
%
% params (threshold,maxiter,beta,decayrate,earlystop)
% model (alpha,TermFactors,AuthFactors,DocMixing)

[mock unused] = process_options(varargin, 'mock', false);

% Compute dimensions with respect to the test set.
Nw = rows(N);
Np = cols(N);
Na = rows(A);

[Ni,Nj] = find(N); Nlen = length(Ni);
[Ai,Aj] = find(A); Alen = length(Ai);

epsilon = lparams.threshold + 1;
iter = 1;
ll.train = [ -Inf ];
ll.temper = [ -Inf ];
newmodel = model;

fprintf(1,'Iteration %2d (train = %.3f, temper = %.3f)\n',...
        0, ll.train(1), ll.temper(1));
while iter <= lparams.maxiter
  fprintf(1,'Iteration %2d ',iter);

  if mock
    iter = iter + 1;
    gain = rand;
    newmodel = model;
    fprintf(1, '\n');
    return;
  end

  % p(zk|ti,dj) -- each column is a distribution
  LatentByDocs = model.TermFactors(:,Ni) .* model.DocMixing(:,Nj);
  LatentByDocs = LatentByDocs.^model.beta;
  LatentByDocs = safe_normalize_cols(LatentByDocs);

  % p(zk|ai,dj) -- each column is a distribution
  LatentByCits = model.AuthFactors(:,Ai) .* model.DocMixing(:,Aj);
  LatentByCits = LatentByCits.^model.beta;
  LatentByCits = safe_normalize_cols(LatentByCits);

  for k = 1:K
    X = N .* sparse(Ni,Nj,LatentByDocs(k,:),Nw,Np);
    Y = A .* sparse(Ai,Aj,LatentByCits(k,:),Na,Np);
    if not(lparams.fold)
      newmodel.TermFactors(k,:) = sum(X,2);
      newmodel.AuthFactors(k,:) = sum(Y,2);
    end
    alpha = model.alpha;
    newmodel.DocMixing(k,:) = alpha .* sum(X,1) + (1-alpha) .* sum(Y,1);
  end
  if not(lparams.fold)
    newmodel.TermFactors = safe_normalize_rows(newmodel.TermFactors);
    newmodel.AuthFactors = safe_normalize_rows(newmodel.AuthFactors);
  end
  newmodel.DocMixing = safe_normalize_cols(newmodel.DocMixing);

  t1 = ch_log_likelihood(N,A,newmodel);
  epsilon = t1-ll.train(end);

  t2 = ch_log_likelihood(Nt,At,newmodel);
  gain = t2-ll.temper(end);

  fprintf(1,'(train = %.3f, temper = %.3f)\n',t1,t2);

  if gain < 0 && lparams.earlystop
    % If the performance on a hold out data set is worse than
    % the last iteration, stop (EM with early stopping).
    return;
  end

  if epsilon < lparams.threshold && not(lparams.earlystop)
    % If the reduction in training loss is small, we allow early stopping
    % even if it isn't explicited specified in lparams. This speeds up
    % the final, non-tempered, EM phase.
    return;
  end

  model = newmodel;
  ll.train = [ll.train t1];
  ll.temper = [ll.temper t2];
  iter = iter + 1;
end
iter = iter - 1;

function debug_entities_with_no_information(N,A)

fprintf(1, 'There are %d documents with no word info\n', ...
        length(find(sum(N,1)==0)));
fprintf(1, 'There are %d documents with no author info\n', ...
        length(find(sum(A,1)==0)));
fprintf(1, 'There are %d documents with neither word nor author info\n', ...
        length(intersect(find(sum(N,1)==0),find(sum(A,1)==0))));
