function [obj cpu models params] = exp_alpha_matters(splits, basepar, alphas, varargin)
% EXP_ALPHA_MATTERS - Run the three-factor model for different alpha's
%
% Arguments:
%
% splits: Cell array of splits
% basepar: Basic parameters shared by all runs. Instance of @glmparams.
% alpha: Vector of weights in favour of data matrix X over data matrix Y.
%
% Return Values:
%
% models: s x a cell array, whose entries are the final model produced
%         for that (split,params) pair.

% Author: ajit@cs.cmu.edu (Ajit Singh)

check_type(splits, 'cell');
check_type(basepar, 'glmparams');
assert(all(isbetween(alphas, 0.0, 1.0, 'closed')), ...
       'Expected all between 0 and 1. Got alphas = [%s]', ...
       sprintf(' %.1f ', alphas));

init_model = random_model3f(splits{1}.X_train, splits{1}.Y_train, basepar);
assert(not(varargin_contains('init_model', varargin{:})));

params = cell(1, length(alphas));
for i = 1:length(alphas)
  params{i} = basepar;
  params{i}.alpha = alphas(i);
end

init_random_seed;
start_time_msg = sprintf('Started %s\n', datestr(now,'dd-mmm-yyyy HH:MM:SS'));
fprintf(1, '%s', start_time_msg);

[obj cpu models] = run_glm_params(splits, params, 'init_model', init_model, ...
                                  varargin{:});

fprintf(1, '%s', start_time_msg);
fprintf(1, 'Finished %s\n', datestr(now,'dd-mmm-yyyy HH:MM:SS'));

models = strip_all_but_last_model(models);

function newmodels = strip_all_but_last_model(models)
% STRIP_ALL_BUT_LAST_MODEL - Instead of storing a vector of models at each
% iteration of each split-parameter pair, just store the final model.
%
% Arguments:
%
% models: An s x p x m cell tensor, where the last dimension is the model at
%         each iteration of the glm call. (Technically, the last dimension is
%         a list of models, so we can index it like an array).
%
% Return Values:
%
% newmodels: An s x p x 1 cell tensor, where the last dimension now contains
%            only the final model.

newmodels = cell(rows(models), cols(models));
for s = 1:rows(models)
  for p = 1:cols(models)
    newmodels{s, p} = models{s, p}(end);
  end
end
