function [cpu models splits alphas desc] = exp_compare_ranking(splits, basepar, alphas, varargin)
% EXP_COMPARE_RANKING - Compare different multiple matrix models at ranking
%
% [cpu models desc] = exp_compare_ranking(splits, basepar, alpha, varargin)
% Executes several different two matrix models:
% + CMF-Identity: Collective matrix factorization using identity links.
% + CMF-Logistic: Collective matrix factorization using logistic links.
% + pLSI-pHITS: Joint pLSI-pHITS models (Cohn & Hofmann 2000)
% Let nmethods denote the number of different models computed.
%
% Arguments:
%
% splits: Cell array of splits (length s)
% basepar: Basic parameters shared by all runs, an instance of @glmparams.
% params: Cell array of parameters, one per glm approach (length p)
% alphas: Vector of weights in favour of data matrix X over data matrix Y
%         (length a)
%
% Return Values:
%
% cpu: nmethods x s x a tensor whose entry is the CPU time required
%      to learn the model.
% models: nmethods x s x a cell tensor whose entry is the model learned.
%         This may be an instance of @model3f, or in the case of pLSI just
%         a structure containing TermFactors, AuthFactors, and DocMixing.
% alphas: Same as input argument.
% desc: 1 x nmethods cell tensor whose entry is a string desc of the method.

% Author: ajit@cs.cmu.edu (Ajit Singh)

check_type(splits, 'cell');
check_type(basepar, 'glmparams');
assert(all(isbetween(alphas, 0.0, 1.0, 'closed')), ...
       'Expected all between 0 and 1. Got alphas = [%s]', ...
       sprintf(' %.1f ', alphas));

init_model = random_model3f(splits{1}.X_train, splits{1}.Y_train, basepar);
assert(not(varargin_contains('init_model', varargin{:})));

nmethods = 3;
cpu = cell(1,nmethods);
models = cell(1, nmethods);
desc = cell(1, nmethods);

desc{1} = 'CMF-Identity';
param = change_link(basepar, 'X', 'gaussian');
param = change_link(basepar, 'Y', 'gaussian');
[ignore cpu_l models_l] = exp_alpha_matters(splits, param, alphas, varargin{:});
cpu_l_m = zeros(length(splits), length(alphas));
for s = 1:length(splits)
  for a = 1:length(alphas)
    cpu_l_m(s,a) = cpu_l{s,a}(end);
  end
end
cpu{1} = cpu_l_m;
models{1} = models_l;

desc{2} = 'CMF-Logistic';
param = change_link(basepar, 'X', 'logistic');
param = change_link(basepar, 'Y', 'logistic');
[ignore cpu_l models_l] = exp_alpha_matters(splits, param, alphas, varargin{:});
cpu_l_m = zeros(length(splits), length(alphas));
for s = 1:length(splits)
  for a = 1:length(alphas)
    cpu_l_m(s,a) = cpu_l{s,a}(end);
  end
end
cpu{2} = cpu_l_m;
models{2} = models_l;

% $$$ desc{2} = 'CMF-Poisson-Logistic';
% $$$ param = change_link(basepar, 'X', 'poisson');
% $$$ param = change_link(basepar, 'Y', 'logistic');
% $$$ [ignore cpu_l models_l] = exp_alpha_matters(splits, param, alphas, varargin{:});
% $$$ cpu_l_m = zeros(length(splits), length(alphas));
% $$$ for s = 1:length(splits)
% $$$   for a = 1:length(alphas)
% $$$     cpu_l_m(s,a) = cpu_l{s,a}(end);
% $$$   end
% $$$ end
% $$$ cpu{2} = cpu_l_m;
% $$$ models{2} = models_l;

desc{3} = 'pLSI-pHITS';
param = basepar;
[cpu_l models_l] = exp_ch_alpha_matters(splits, param, alphas, varargin{:});
cpu{3} = cell2mat(cpu_l);
models{3} = models_l;

splits = strip_splits(splits);

function splits = strip_splits(splits)

check_type(splits, 'cell');
msg = 'removed';
for s = 1:length(splits)
  splits{s}.Y_train = msg;
  splits{s}.ytest = msg;
  splits{s}.eyi = msg;
  splits{s}.eyj = msg;
  splits{s}.wyf = msg;
  splits{s}.wy_context = msg;
  splits{s}.eval_y = msg;
end
