function [obj cpu models params] = exp_newton(splits, basepar, varargin)
% EXP_ALPHA_MATTERS - Run the three-factor model for different alpha's
%
% Arguments:
%
% splits: Cell array of splits
% basepar: Basic parameters shared by all runs. Instance of @glmparams.
% alpha: Vector of weights in favour of data matrix X over data matrix Y.
%
% Return Values:

% Author: ajit@cs.cmu.edu (Ajit Singh)

check_type(splits, 'cell');
check_type(basepar, 'glmparams');
WarnIf(basepar.maxiter > 20, 'Slow: will use %d gradient steps', basepar.maxiter^2);

init_model = random_model3f(splits{1}.X_train, splits{1}.Y_train, basepar);
assert(not(varargin_contains('init_model', varargin{:})));
assert(not(varargin_contains('gradient', varargin{:})));

[grad_maxiter unused] = ...
    process_options(varargin, 'grad_maxiter', basepar.maxiter^2);

init_random_seed;
start_time_msg = sprintf('Started %s\n', datestr(now,'dd-mmm-yyyy HH:MM:SS'));
fprintf(1, '%s', start_time_msg);

[obj.newton cpu.newton models.newton] = ...
    run_glm_params(splits, {basepar}, 'init_model', init_model, varargin{:});

basepar = force_iterations(basepar, grad_maxiter);
[obj.gradient cpu.gradient models.gradient] = ...
    run_glm_params(splits, {basepar}, 'init_model', init_model, ...
                   'gradient', true, varargin{:});

fprintf(1, '%s', start_time_msg);
fprintf(1, 'Finished %s\n', datestr(now,'dd-mmm-yyyy HH:MM:SS'));

models.newton = strip_all_but_last_model(models.newton);
models.gradient = strip_all_but_last_model(models.gradient);
params = {basepar};

function newmodels = strip_all_but_last_model(models)
% STRIP_ALL_BUT_LAST_MODEL - Instead of storing a vector of models at each
% iteration of each split-parameter pair, just store the final model.
%
% Arguments:
%
% models: An s x p x m cell tensor, where the last dimension is the model at
%         each iteration of the glm call. (Technically, the last dimension is
%         a list of models, so we can index it like an array).
%
% Return Values:
%
% newmodels: An s x p x 1 cell tensor, where the last dimension now contains
%            only the final model.

newmodels = cell(rows(models), cols(models));
for s = 1:rows(models)
  for p = 1:cols(models)
    newmodels{s, p} = models{s, p}(end);
  end
end
