function [obj cpu models params] = ...
    exp_stochastic(splits, basepar, batches, ...
                   niters_newton, niters_stochastic, varargin)
% EXP_STOCHASTIC - Compare Newton vs. Stochastic on the 3-factor model.
%
% Arguments:
%
% splits: Cell array of splits (1 x s)
% basepar: Basic parameters (link/loss/debugging) shared by all runs. @glmparams
% batches: Vector of batch sizes for stochastic gradient (1 x b)
% niters_newton: Number of iterations of Newton to run.
% niters_stochastic: Number of iterations of stochastic gradient to run.
%
% Return Values:
%
% Let p = 1 + length(batches) be the number of parameter settings used.
%
% obj: s x p cell matrix where each cell is a vector of training errors at
%      each iteration.
% loss: s x p cell matrix where each cell is a structure with fields x and y.
%       The first contains a vector of Bregman losses on the test entries of
%       x. The latter contains a vector of Bregman losses on test entries of y
% err: s x p cell matrix. Like loss except it uses the error metric given
%      by the split, which is usually not Bregman (e.g., MAE)
% cputime: s x p cell matrix where each cell is a vector of cpu time (in sec)
%          required to finish the first k iterations.
% models: s x p cell matrix whre each cell is a cell array of models created
%         at each iteration (as @model3f).

% Author: Ajit Singh (ajit@cs.cmu.edu)

error(nargchk(5, Inf, nargin));

check_splits(splits);
check_type(basepar, 'glmparams');
assert(not(isempty(batches)));
assert(niters_newton > 0 || niters_stochastic > 0, ...
       'Disabled both Newton and Stochastic runs. No point in running me.');

init_model = random_model3f(splits{1}.X_train, splits{1}.Y_train, basepar);
assert(all(cellfun(@(argname)(not(strcmpi(argname, 'init_model'))), varargin)), ...
       'init_model defined earlier in calling sequence');

[min_scaling_factor unused] = ...
    process_options(varargin, 'min_scaling_factor', ones(1, length(batches)));

if niters_stochastic <= 0
  batches = [ ];
  min_scaling_factor = [ ];
end
check_length(min_scaling_factor, length(batches));

if niters_newton > 0
  nmethods = 1 + length(batches);
  params = cell(1, nmethods);
  params{1} = disable_stochastic(basepar);
  params{1} = force_iterations(params{1}, niters_newton);
  for i = 2:nmethods
    params{i} = enable_stochastic(basepar, batches(i-1));
    params{i} = force_iterations(params{i}, niters_stochastic);
    params{i}.min_scaling_factor = min_scaling_factor(i-1);
  end
else
  nmethods = length(batches);
  params = cell(1, nmethods);
  for i = 1:nmethods
    params{i} = enable_stochastic(basepar, batches(i));
    params{i} = force_iterations(params{i}, niters_stochastic);
    params{i}.min_scaling_factor = min_scaling_factor(i);
  end
end

init_random_seed;
start_time_msg = sprintf('Started %s\n', datestr(now,'dd-mmm-yyyy HH:MM:SS'));
fprintf(1, '%s', start_time_msg);

[obj cpu models] = run_glm_params(splits, params, ...
                                  'init_model', init_model, varargin{:});

fprintf(1, '%s', start_time_msg);
fprintf(1, 'Finished %s\n', datestr(now,'dd-mmm-yyyy HH:MM:SS'));

