function [U V obj debug] = glm1(X, W, k, params)
% glm1 - Given one matrix X (m x n) factor it into X = UV' where the error
%        between the matrix and its reconstruction is a Bregman divergence.
%
% [U, V, obj] = glm1(X, W, k, params)
% Generalization of matrix factorization to regularized generalized linear
% models. Given a matrix X (m x n) generate factors U and V such that
% the sum of Bregman divergences D_F(X||UV') + D_G(0||U) + D_H(0||V) is small.
%
% The weight field W takes either a matrix or a weight function. The latter
% is generated by the @class/data* functions. For example, see @movie/data.m
% The test_train_splits field is a function that generates splits, each of
% which contains a weight function (the wxf and wyf fields).
%
% While this function is just a special case of glm.m, this separate
% implementation allows easy experimentation.
%
% Arguments
%   X:  An m x n sparse matrix.
%   W:  A weight function or weight matrix.
%   k:  Number of dimensions to embed each matrix into.
%   alpha: The relative weight favouring the reconstruction of X over Y
%          (in [0,1]). Larger values of alpha mean reconstructing X is
%          more important.
%   params: Parameters defining the factorization (links, losses)
%
% Returns
%   U: An m x (k+1) factor matrix. The last column is constant 1's
%   V: An n x (k+1) factor matrix. The last column are row bias terms.
%   obj: The loss function at each iteration. The first value is
%        the objective given random starting parameters.
%   debug: Debugging term, see code.

% Author: ajit@cs.cmu.edu (Ajit Singh)

error(nargchk(4, 4, nargin));

debug.trace_bound = [ ];
debug.trace_norm = [ ];

[m, n] = size(X);

WarnIf(params.stochastic_batch_size > min(size(X)), ...
       ['Batch size > smallest dim of data matrix (%d vs. %d).\n' ...
        'When required, we will shrink the batch size to %d'], ...
       params.stochastic_batch_size, min(size(X)), min(size(X)));

glmvars.k = k;
if params.add_bias_term
  glmvars.Gi = params.scale_g * eye(k+1);
  glmvars.Gi(end, end) = 0;
  glmvars.Hi = params.scale_h * eye(k+1);
else
  glmvars.Gi = params.scale_g * eye(k);
  glmvars.Hi = params.scale_h * eye(k);
end

if isa(W, 'function_handle')
  glmvars.wx_func = W;
  glmvars.hasweights_x = false;
else
  check_dims(W, rows(X), cols(X), 'W');
  glmvars.Wx = W;
  glmvars.hasweights_x = true;
  if not(isnumeric(W))
    warning('RFA:Performance', ...
            'W is of type %s, which will make w .* err very slow', class(W));
  end
end

if params.add_bias_term
  % Bias is encoded in the factors by fixing the last column of U to 1's,
  % so that the last column of V is the bias term. This induces a per column
  % term in the model (for a user x movie matrix, this is a movie offset)
  U = [normalize_rows(rand(m, k)) ones(m, 1)];
  V = normalize_rows(rand(n, k+1));
else
  U = normalize_rows(rand(m, k));
  V = normalize_rows(rand(n, k));
end

obj = [ loss(X, U, V, params, glmvars) ];
fprintf(1, 'Iteration 0 (random parameters): obj = %g\n', obj(1));

for iter = 1:params.maxiter
  % Compute trace norm and bound from MMMF regularization. Assumes
  % param links are g(x) = x and h(x) = x.
% $$$   debug.trace_bound = [debug.trace_bound ...
% $$$                       1/2 * (norm(U, 'fro')^2 + norm(V, 'fro')^2)];
% $$$   [Af,Sf,Bf] = svd(U*V');
% $$$   debug.trace_norm = [debug.trace_norm trace(Sf)];

  normchange = repmat(NaN, 1, 2);
  [Un ignore normchange(1)] = newton_step_u(X, U, V, iter, params, glmvars);
  [Vn ignore normchange(2)] = newton_step_v(X, Un, V, iter, params, glmvars);
  epsilon = loss(X, Un, Vn, params, glmvars) - obj(end);

  if epsilon > 0
    warning('RFA:BadResult', 'Freaky, epsilon = %e\n', epsilon);
    if not(params.stochastic_opt)
      break;
    end
  end

  if isfinite(epsilon)
    obj = [obj obj(end) + epsilon];
    U = Un;
    V = Vn;
  else
    warning('New parameters have infinite loss, not using the update');
  end

  if mod(iter,1) == 0
    fprintf(1, ['Iteration %d: eps = %g, obj = %g, nc = %.1e/%.1e, '...
                'oc = %.3f\n'], iter, epsilon, obj(end), ...
            normchange(1), normchange(2), abs(epsilon/obj(end)));
  end

  if abs(epsilon/obj(end)) < params.obj_threshold && iter >= params.miniter
    break
  end
end

function [U, delta, change] = newton_step_u(X, U, V, iter, params, glmvars)

  [m, n] = size(X);
  k = glmvars.k;
  if params.add_bias_term
    check_dims(U, m, k+1, 'U');
    check_dims(V, n, k+1, 'V');
    check_dims(glmvars.Gi, k+1, k+1, 'Gi');
  else
    check_dims(U, m, k, 'U');
    check_dims(V, n, k, 'V');
    check_dims(glmvars.Gi, k, k, 'Gi');
  end

  Uold = U;
  Gi = glmvars.Gi;

  condition = repmat(NaN, 1, m);
  stepsize = repmat(NaN, 1, m);
  unchanged = 0;
  delta = 0;
  D = spdiags(repmat(1, n, 1), 0, n, n);
  didx = find(D);
  row_idx = repmat(NaN, 1, n);
  column_idx = 1:n;

  Vorig = V;
  power = 1;
  for i = 1:m
    power = TextProgressMessage(i, m, power, 'Updating U: ', params.show_iter);

    if params.stochastic_opt
      s = sample(1:rows(V), min(params.stochastic_batch_size, rows(V)));
      V = Vorig(s,:);
    else
      s = 1:rows(Vorig);
      V = Vorig;
    end

    % Compute descent direction.
    x = X(i,s);
    u = U(i,:);
    uv = u*V';
    if glmvars.hasweights_x
      w = glmvars.Wx(i,s);
    else
      row_idx(:) = i;
      w = glmvars.wx_func(x, repmat(i, 1, length(s)), s);
      %w = glmvars.wx_func(x, row_idx, column_idx);
      %w = glmvars.wx_func(x, [ ], [ ]);
    end
    if all(w == 0)
      LogIf(params.debug && params.debug_zeroweights, ...
            'U(%d): All weights w == 0\n', i);
    end
    expectx = params.f(uv);
    err = (x - expectx);
    raw_deriv = params.f_prime(uv);
    deriv = replace_zero_deriv(raw_deriv, err, sprintf('U(%d)', i), params);
    scaled_err = err ./ deriv;
    D = spdiags((w .* raw_deriv)', 0, rows(V), rows(V));
    %D(didx) = w .* raw_deriv;

    covar = V' * D * V + Gi;
    Ai = params.inv_g(u);

    % If the covariance cannot be inverted safely, then do not update the
    % parameters. One way to improve the condition number is to increase
    % params.scale, which increases the amount of regularization.
    if params.check_condition_number
      cond = safe_cond(covar);
      condition(i) = cond;
      if cond > params.ill_conditioned_threshold || not(isfinite(cond))
        LogIf(params.debug && params.debug_cond,...
              'Cancelling update of U(%d). Condition num = %f\n', i, cond);
        unchanged = unchanged + 1;
        continue;
      end
    end

    % Armijo line search
    q_old = (w.* -err) * V + Ai;
    step = q_old / covar;
    loss_old = xloss_natural(x, w, uv, u, V, params, glmvars);
    loss_new = loss_old;

    if params.stochastic_opt
      eta = 1/iter;
    else
      eta = 1;
    end
    sufficient = false;
    while eta > params.mineta && not(sufficient)
      adjx = (uv + eta .* scaled_err);
      u_new = (adjx * D * V + u*Gi - eta .* Ai) / covar;
      if params.add_bias_term
        u_new = [u_new(1:k) 1];
      end
      loss_new = xloss(x, w, u_new, V, params, glmvars);
      sufficient = (loss_new - loss_old <= params.c * eta * step * q_old');
      if sufficient || params.stochastic_opt
        U(i,:) = u_new;
        stepsize(i) = eta;
        delta = delta + (loss_new - loss_old);
        break;
      end
      if not(params.linesearch) || iter < params.min_armijo_iter
        break
      end
      eta = params.rho * eta;
    end
    if not(sufficient)
      LogIf(params.debug && params.debug_decrease,...
            'Cancelling update of U(%d). No good stepsize found\n', i);
      unchanged = unchanged + 1;
    end
  end

  LogIf(params.check_condition_number && ...
        params.debug_summary_condition_number, ...
        'Condition U: %g %g %g\n', ...
        min(condition), median(condition), max(condition));

  stepsize = stepsize(isfinite(stepsize));
  if isempty(stepsize)
    stepsize = [0];
  end
  LogIf(params.debug && params.debug_rowstats, ...
        ['Num unchanged U: %d/%d\t(eta: min/mean/max = %.2g/%.2g/%.2g) '...
         '(delta = %g)\n'], ...
        unchanged, m, min(stepsize), mean(stepsize), max(stepsize), delta);

  change = sum(sum((U - Uold).^2))/numel(U);
  LogIf(params.debug && params.debug_normchange, ...
        'Mean L2 normchange for U: %f\n', change);

function [V, delta, change] = newton_step_v(X, U, V, iter, params, glmvars)
  [m, n] = size(X);
  k = glmvars.k;
  if params.add_bias_term
    check_dims(U, m, k+1, 'U');
    check_dims(V, n, k+1, 'V');
    check_dims(glmvars.Hi, k+1, k+1, 'Hi');
  else
    check_dims(U, m, k, 'U');
    check_dims(V, n, k, 'V');
    check_dims(glmvars.Hi, k, k, 'Hi');
  end

  Vold = V;
  Hi = glmvars.Hi;
  condition = repmat(NaN, 1, n);
  stepsize = repmat(NaN, 1, n);
  unchanged = 0;
  delta = 0;
  Dx = spdiags(ones(m, 1), 0, m, m);
  didx_x = find(Dx);
  row_idx_x = 1:m;
  column_idx_x = repmat(NaN, 1, m);

  Uorig = U;
  power = 1;
  for i = 1:n
    power = TextProgressMessage(i, n, power, 'Updating V: ', params.show_iter);

    if params.stochastic_opt
      s = sample(1:rows(Uorig), min(params.stochastic_batch_size, rows(Uorig)));
      U = Uorig(s,:);
    else
      s = 1:rows(Uorig);
      U = Uorig;
    end

    % Compute descent direction
    x = X(s,i);
    v = V(i,:);
    uv = U*v';
    if glmvars.hasweights_x
      wx = Wx(s,i);
    else
      column_idx_x(:) = i;
      wx = glmvars.wx_func(x, s, repmat(i, 1, length(s)));
      %wx = glmvars.wx_func(x, [ ], [ ]);
    end
    expectx = params.f(uv);
    errx = (x - expectx)';
    raw_derivx = params.f_prime(uv);
    derivx = replace_zero_deriv(raw_derivx, errx', sprintf('V(%d)', i), params);
    scaled_errx = errx ./ derivx';
    %Dx(didx_x) = wx .* raw_derivx;
    Dx = spdiags(wx .* raw_derivx, 0, rows(U), rows(U));

    covar = U' * Dx * U + Hi;
    Bi = params.inv_i(v);

    if params.check_condition_number
      cond = safe_cond(covar);
      condition(i) = cond;
      if cond > params.ill_conditioned_threshold || not(isfinite(cond))
        LogIf(params.debug && params.debug_cond, ...
              'Cancelling update of V(%d). Condition num = %f\n', i, cond);
        unchanged = unchanged + 1;
        continue;
      end
    end

    % Armijo line search
    q_old = -(wx' .* errx) * U + Bi;
    step = q_old / covar;
    loss_old = xloss_natural(x, wx, uv, U, v, params, glmvars);
    loss_new = loss_old;

    if params.stochastic_opt
      eta = 1/iter;
    else
      eta = 1;
    end
    sufficient = false;
    while eta > params.mineta && not(sufficient)
      adjx = (uv' + eta .* scaled_errx);
      adjusted = (adjx * Dx * U) + v*Hi - eta.*Bi;
      v_new = adjusted / covar;
      loss_new = xloss(x, wx, U, v_new, params, glmvars);
      sufficient = (loss_new - loss_old <= params.c * eta * step * q_old');
      if sufficient || params.stochastic_opt
        V(i,:) = v_new;
        stepsize(i) = eta;
        delta = delta + (loss_new - loss_old);
        break;
      end
      if not(params.linesearch) || iter < params.min_armijo_iter
        break;
      end
      eta = params.rho * eta;
    end
    if not(sufficient)
      LogIf(params.debug && params.debug_decrease,...
            'Cancelling update of V(%d). No good stepsize found.\n', i);
      unchanged = unchanged + 1;
    end
  end

  LogIf(params.check_condition_number && ...
        params.debug_summary_condition_number, ...
        'Condition V: %g %g %g\n', ...
        min(condition), median(condition), max(condition));

  stepsize = stepsize(isfinite(stepsize));
  if isempty(stepsize)
    stepsize = [0];
  end
  LogIf(params.debug && params.debug_rowstats, ...
        ['Num unchanged V: %d/%d\t(eta: min/mean/max = %.2g/%.2g/%.2g) ',...
         '(delta = %g)\n'], ...
        unchanged, n, min(stepsize), mean(stepsize), max(stepsize), delta);

  change = sum(sum((V - Vold).^2))/numel(V);
  LogIf(params.debug && params.debug_normchange, ...
        'Mean L2 normchange for V: %f\n', change);

function obj = xloss_natural(X, W, T, U, V, params, glmvars)
% xloss_natural -- Compute the reconstruction loss of the X matrix, given
% the factors and the natural parameters.
%
% obj = xloss_natural(X,W,T,U,V,params)
% Computes the reconstruction error of X by f(T). The data, weights, and
% parameters can be either vectors or matrices.
%
% Arguments:
%
% X: Data matrix (m x n)
% W: Weight matrix (m x n)
% T: Natural parameters (m x n)
% U: Row factors (m x k)
% V: Column factors (n x k)
% params: Optimization parameters
%
% Return Values:
%
% obj: Reconstruction error

  loss = full(params.loss_f(X, W, T, params.f));
  if params.add_bias_term
    % The constant column of U are not real parameters, so we shouldn't
    % pay a regularization penalty for them.
    k = glmvars.k;
    U = U(:,1:k);
  end
  reg_u = params.scale_g * sum(sum(U.^2))/(2*numel(U));
  reg_v = params.scale_h * sum(sum(V.^2))/(2*numel(V));
  obj = loss + reg_u + reg_v;
  %fprintf(1, 'X: loss = %g, reg_u = %g, reg_v = %g\n', loss, reg_u, reg_v);

function obj = xloss(X, W, U, V, params, glmvars)
% loss_x -- Compute the reconstruction loss of the X matrix.
%
% obj = loss_x(X,W,U,V,params)
% Computes the weighted reconstruction error of X as f(UV') where
% U = g(A) and V = h(B). The functions works regardless of whether
% the data and the reconstructions are vectors or matrices.
%
% When params.alpha == 0, this function always returns 0.
%
% Arguments:
%
% X: Matrix we want to approximate
% W: Reconstruction weights. Precondition: size(X) == size(W)
% U: Row factors for X. Precondition: rows(X) == rows(U)
% V: Col factors for X. Precondition: cols(X) == rows(V)
% params: Optimization parameters. Precondition: isfield(params,'alpha'),
%         isfield(params,'scale_g'), isfield(params,'scale_h'),
%         isfield(params,'f')
%
% Return Values:
%
% obj: Reconstruction error D_F(UV'|X,W) + D_G(0|U) + D_H(0|V). Note,
%      the error has not been scaled by alpha.

assert(rows(X) == rows(U), ...
       'RFA:InvalidArgument', 'rows(X) ~= rows(U) (%d vs. %d)', ...
       rows(X), rows(U));

assert(cols(X) == rows(V), ...
       'RFA:InvalidArgument', 'cols(X) ~= rows(V) (%d vs. %d)', ...
       cols(X), rows(V));

check_dims(X, rows(W), cols(W), 'X');
obj = xloss_natural(X, W, U*V', U, V, params, glmvars);

function obj = loss(X, U, V, params, glmvars)
% loss -- Compute the overall loss of the three entity-type model.
%
% obj = loss(X,Y,U,V,Z,params)
% Compute the overall loss of the three entity-type model:
%    alpha * [D_F(UV'|X) + D_G(0|U) + D_H(0|V)] +
%    (1-alpha) * [D_F(VZ'|Y) + D_H(0|V) + D_I(0|Z)]
%
% Implementation Details:
%
% Designed to compute the loss of each row of X and each column of
% Y, which precludes creating a large dense weight matrix.
%
% Arguments:
%
% X: An m x n matrix.
% Y: An n x r matrix.
% U: An m x k matrix of factors.
% V: An n x k matrix of shared factors.
% Z: An r x k matrix of factors.
% params: Optimization parameters. Precondition: isfield(params,'alpha')
%
% Return Values:
%
% obj: Reconstruction loss.

obj = 0;
m = rows(X);
n = cols(X);
row_idx = repmat(NaN, 1, n);
column_idx = 1:n;
for i = 1:m
  x = X(i,:);
  if glmvars.hasweights_x
    w = glmvars.Wx(i,:);
  else
    row_idx(:) = i;
    w = glmvars.wx_func(x, row_idx, column_idx);
  end
  u = U(i,:);
  obj = obj + xloss_natural(x, w, u*V', u, V, params, glmvars);
end

function [deriv, zeroderiv] = replace_zero_deriv(deriv, err, msg, params)
% Replace all zero derivatives with a non-zero value.
%
% Arguments:
%
% deriv: Vector of derivatives
% err: Vectors of residual errors
% msg: Prefix of a debugging message recording the number of zero derivatives.
% params: Optimization parameters (only debugging parameters are
% used).
%
% Return Values:
%
% deriv: Vector of derivatives, with no zero entries.
% zeroderiv: Boolean indicating whether or not any entries were replaced.

zeroderiv = (deriv == 0 & err ~= 0);
if any(zeroderiv)
  LogIf(params.debug && params.debug_deriv,...
        '%s: There are %d zero derivatives with nz err\n', msg, sum(zeroderiv));
end
deriv(deriv == 0) = 1;

function k = safe_cond(M)
% Compute the condition number of a potentially non-finite matrix.
%
% Arguments:
%
% M: The matrix whose condition number we want.
%
% Return Values;
%
% k: If any entry of M is not finite, NaN. Otherwise k = cond(M).

if not(all(isfinite(M)))
  k = NaN;
else
  k = cond(M);
end

