function [model obj debug] = glm4f(X, Wx, wx_con, Y, Wy, wy_con, ...
                                   Q, Wq, wq_con, ...
                                   params, split, varargin)
% GLM - Factors two matrices X = UV' and Y = VZ' where the V factors is shared.
%
% [model obj debug] = glm(X, Wx, Y, Wy, params, varargin)
% Given two matrices X (m x n) and Y (n x r) this algorithms factors them
% into X = UV' and Y = VZ' where error between the matrices and their
% reconstructions are Bregman divergences.
%
% Arguments:
%
% X: Data matrix (m x n)
% Y: Data matrix (n x r)
% Wx: Weights for X, either as an explicit matrix or implicitly as a function.
% Wy: Weights for Y, either as an explicit matrix or implicitly as a function.
% params: Optimization and model parameters. Instance of @glmparams.
%
% Optional Arguments:
%
% gradient: If true, optimize using gradient descent instead of Newton.
%           [default: false]
% mock: Mock the parameter updates, return random parameters. [default: false]
% mock_loss: Mock the reconstruction error. [default: false]
% init_model: Used the given @model3f as starting parameters. [default: none]
%
% Return Values:
%
% model: The learned model. Instance of @model3f.
% obj: The training error after each update. The first value is the training
%      error using random parameters (or init_model).
% debug: Anything that peers into the internals of the optimization. Right
%        now the only fields are 'cputime', the time taken to finish the first
%        k iterations; and 'model', a vector of models generated at each
%        iteration

% Author: ajit@cs.cmu.edu (Ajit Singh)

error(nargchk(11, Inf, nargin));

[m, n] = size(X);
[q, r] = size(Y);
[c, s] = size(Q);

% HACK: If X is sparse then accessing rows is very expensive (column-major
% storage is used). This can add > 30% to the cost of each iteration for
% large data sets. Where we need a row of X, X(i,:), it can be replaced by
% a column of Xt, Xt(:,i)'. Likewise for Y.
Xt = X';
Yt = Y';
Qt = Q';

% HACK: Field references for objects (subsref) are much slower than field
% references for structures. To avoid subsref being the dominant cost in
% many utility functions, such as sample_factor, cast @glmparams -> struct.
params = struct(params);

alpha = params.alpha;
glmvars.alpha = params.alpha;
k = params.dims;
glmvars.k = params.dims;

assert(n == q, 'Dimensions of X and Y incompatable (%d x %d vs. %d x %d)', ...
       rows(X), cols(X), rows(Y), cols(Y));

assert(n == c, 'Dimensions of X and Z incompatable (%d x %d vs. %d x %d)', ...
       rows(X), cols(X), rows(Q), cols(Q));

assert(length(alpha) == 3, ...
       'Require exactly three data weight, got %d', length(alpha));

assert(all(isbetween(alpha, 0.0, 1.0, 'closed')), ...
       'alpha must be in [0,1] (is %g %g)', alpha);

assert(any(alpha ~= 0), 'All the data weights are zero');

WarnIf(params.stochastic_opt && params.stochastic_batch_size > min(size(X)), ...
       ['Batch size > smallest dim of data matrix X (%d vs. %d).\n' ...
        'When required, we will shrink the batch size to %d'], ...
       params.stochastic_batch_size, min(size(X)), min(size(X)));

WarnIf(params.stochastic_opt && params.stochastic_batch_size > min(size(Y)), ...
       ['Batch size > smallest dim of data matrix Y (%d vs. %d).\n' ...
        'When required, we will shrink the batch size to %d'], ...
       params.stochastic_batch_size, min(size(Y)), min(size(Y)));

[glmvars.mock unused] = process_options(varargin, 'mock', false);
[glmvars.mock_loss unused] = process_options(varargin, 'mock_loss', false);
[glmvars.mock_eval unused] = process_options(varargin, 'mock_eval', false);

WarnIf(glmvars.mock, 'Mocking computation of glm.m to test the interface');
WarnIf(glmvars.mock_loss && params.miniter <= 0, ...
       ['Mocking the loss without forcing minimum number of iterations'...
        ' may only run to one iteration']);
WarnIf(glmvars.mock_eval, 'Disable computation of test error per iteration');

[init_model unused] = process_options(varargin, 'init_model', [ ]);
if isa(init_model, 'model4f')
  check_model(init_model, params);
  model = init_model;
else
  model = random_model4f(X, Y, Q, params);
end
[U V Z R] = unwrap(model);

HU = cell(1, m);
HV = cell(1, n);
HZ = cell(1, r);
HR = cell(1, s);

[glmvars.Wx glmvars.wx_func glmvars.hasweights_x] = verify_weights(Wx, X);
[glmvars.Wy glmvars.wy_func glmvars.hasweights_y] = verify_weights(Wy, Y);
[glmvars.Wq glmvars.wq_func glmvars.hasweights_q] = verify_weights(Wq, Q);
glmvars.wx_context = wx_con;
glmvars.wy_context = wy_con;
glmvars.wq_context = wq_con;

[glmvars.gradient unused] = process_options(varargin, 'gradient', false);

LogIf(glmvars.gradient, 'Using the gradient step instead of Newton.\n');
WarnIf(params.add_bias_term, 'Using add_bias_term, experimental');
LogIf(params.no_adjusted_dependent_variates, ...
      'Using no_adjusted_dependent_variates.\n');

debug.cputime = [ 0 ]; % i^th entry is time it took to complete i iterations.
debug.model = [ model ];
obj = [ loss(Xt, Y, Qt, U, V, Z, R, params, glmvars) ];
fprintf(1, 'Iteration 0 (random parameters): obj = %g\n', obj(1));

print_eval = @print_eval_func;

function print_eval_func(model, iter)
if not(glmvars.mock_eval)
  [a b c] = evaluate_error(model, split, params);
  fprintf(1, 'Iteration %d: Test Error: X = %f, Y = %f, Q = %f\n', iter, ...
          a, b, c);
end
end

print_eval_func(model, 0);
for iter = 1:params.maxiter
  normchange = repmat(NaN, 1, 4);
  start_time = cputime;
  [Un HU ignore normchange(1)] = ...
      newton_step_u_dispatch(Xt, U, V, iter, HU, params, glmvars);
  [Zn HZ ignore normchange(2)] = ...
      newton_step_z_dispatch(Y, V, Z, iter, HZ, params, glmvars);
  [Rn HR ignore normchange(4)] = ...
      newton_step_r_dispatch(Q, V, R, iter, HR, params, glmvars);
  [Vn HV ignore normchange(3)] = ...
      newton_step_v_dispatch(X, Yt, Qt, Un, V, Zn, Rn, iter, HV, params, glmvars);
  duration = cputime - start_time;
  epsilon = loss(Xt, Y, Qt, Un, Vn, Zn, Rn, params, glmvars) - obj(end);
  if (epsilon <= 0 && isfinite(epsilon)) || ...
        (iter <= params.miniter && isfinite(epsilon)) || ...
        ((params.stochastic_opt || params.add_bias_term) && isfinite(epsilon))
    obj = [obj obj(end) + epsilon];
    U = Un;
    V = Vn;
    Z = Zn;
    R = Rn;
    model = model4f(Un, Vn, Zn, Rn, params);
    print_eval_func(model, iter);
    debug.model = [ debug.model model ];
    debug.cputime = [ debug.cputime debug.cputime(end) + duration ];
  elseif not(isfinite(epsilon))
    if params.stochastic_opt
      warning('RFA:BadResult', 'Infinite loss...throwing away the update');
    else
      error('RFA:BadResult', 'Infinite loss on non-stochastic descent');
    end
  else
    warning('RFA:BadResult', 'Freaky, epsilon = %e\n', epsilon);
    if not(params.stochastic_opt || params.add_bias_term)
      break;
    end
  end

  oc = abs(epsilon/obj(end));
  if mod(iter,1) == 0
    fprintf(1, ['Iteration %d: eps = %g, obj = %g, '...
                'nc(u/v/z/r) = %.1e/%.1e/%.1e/%.1e '...
                'oc = %.3f, time = %.3f sec\n'], ...
            iter, epsilon, obj(end), ...
            normchange(1), normchange(3), normchange(2), normchange(4), ...
            oc, duration);
  end

  if not(params.stochastic_opt)
    if oc < params.obj_threshold && iter >= params.miniter
      break
    end
  end
end

end % function glm

function [U HU delta change] = ...
    newton_step_u_dispatch(Xt, U, V, iter, HU, params, glmvars)

if glmvars.alpha(1) == 0
  delta = 0;
  change = 0;
  return;
end

k = glmvars.k;
if params.add_bias_term
  error('unsupported');
  pidx = [1:k k+2];
  Gi = diag(params.scale_g * [ones(1,k) 0]);
  [U_tied HU delta change] = newton_step_u(Xt, U(:,pidx), V(:,pidx), ...
                                           Gi, iter, ...
                                           HU, params, glmvars);
  U(:,pidx) = U_tied;
else
  Gi = params.scale_g * eye(k);
  [U HU delta change] = newton_step_u(Xt, U, V, Gi, iter, HU, params, glmvars);
end

end % function newton_step_u_dispatch

function [U, HU, delta, change] = newton_step_u(Xt, U, V, Gi, iter, HU, params, glmvars)

if glmvars.mock
  U = normalize_rows(rand(size(U)));
  delta = 0;
  change = NaN;
  return;
end

[n m] = size(Xt);
k = glmvars.k;
factorname = 'U';
Uold = U;
alpha = glmvars.alpha(1);
condition = repmat(NaN, 1, m);
stepsize = repmat(NaN, 1, m);
unchanged = 0;
delta = 0;
%[D didx] = create_scaling_matrix(V, params);
wx_func = glmvars.wx_func;

if params.inline_standard_weights
  pex = glmvars.wx_context.pex;
  disx = glmvars.wx_context.disx;
  Exclude = glmvars.wx_context.Exclude;
  Exclude_t = glmvars.wx_context.Exclude_trans;
end

Vorig = V;
power = 1;
for i = 1:m
  power = TextProgressMessage(i, m, power, 'Updating U: ', params.show_iter);

  if params.densify_rows
    x = full(Xt(:,i))';
  else
    x = Xt(:,i)';
  end
  u = U(i,:);
  if glmvars.hasweights_x
    w = glmvars.Wx(i,:);
  else
    row_idx = i * ones(1, rows(Vorig));
    col_idx = 1:rows(Vorig);
    if params.inline_standard_weights
      w = NaN * ones(size(x));
      idx = (x ~= 0);
      w(idx) = 1 / pex;
      w(~idx) = disx / pex;
      w(logical(Exclude_t(:,i))) = 0;
    else
      w = wx_func(x, row_idx, col_idx);
    end
  end
  [V s] = sample_factor_weighted(Vorig, w, params);
  x = x(s);
  w = w(s);
  uv = u*V';
  if all(w == 0)
    LogIf(params.debug && params.debug_zeroweights, ...
          'U(%d): All weights w == 0\n', i);
  end
  expectx = params.f(uv);
  err = (x - expectx);
  raw_deriv = params.f_prime(uv);
  deriv = replace_zero_deriv(raw_deriv, err, sprintf('U(%d)', i), params);
  scaled_err = err ./ deriv;
  %D(didx) = w .* raw_deriv;
  D = spdiags((w .* raw_deriv)', 0, rows(V), rows(V));
  Ai = params.inv_g(u);
  if params.add_bias_term
    Ai(k+1:end) = 0;
  end
  eta = initial_steplength(params, glmvars, iter);

  if glmvars.gradient
    % Gradient or stochastic gradient
    grad = alpha * (w .* -err) * V + Ai;
    U(i,:) = U(i,:) - eta * grad;
  elseif params.stochastic_opt
    % Stochastic Newton with an Hessian estimated by an exponentially
    % weighted moving average. The precludes easy use of the adjusted
    % dependent variate form.
    grad = alpha .* (w .* -err) * V + Ai;
    H_new = alpha .* V' * D * V + Gi;
    covar = estimate_stochastic_hessian(H_new, HU{i}, iter);
    invertible = check_covariance(covar, params, factorname, i);
    if not(invertible)
      unchanged = unchanged + 1;
      continue;
    end
    U(i,:) = U(i,:) - eta .* grad / covar;
    HU{i} = covar;
    stepsize(i) = eta;
  else
    % Newton with the adjusted dependent variate update.
    covar = alpha .* V' * D * V + Gi;
    invertible = check_covariance(covar, params, factorname, i);
    if not(invertible)
      unchanged = unchanged + 1;
      continue;
    end

    if params.linesearch && iter >= params.min_armijo_iter
      grad = alpha .* (w.* -err) * V + Ai;
      step = grad / covar;
      loss_old = alpha * xloss_natural(x, w, uv, u, V, params, glmvars);

      sufficient = false;
      while eta > params.mineta && not(sufficient)
        if params.no_adjusted_dependent_variates
          u_new = U(i,:) - eta .* step;
        else
          adjx = alpha .* (uv + eta .* scaled_err);
          u_new = (adjx * D * V + u*Gi - eta .* Ai) / covar;
        end
        loss_new = alpha * xloss(x, w, u_new, V, params, glmvars);
        sufficient = (loss_new - loss_old <= params.c * eta * step * grad');
        if sufficient
          U(i,:) = u_new;
          stepsize(i) = eta;
          delta = delta + (loss_new - loss_old);
          break;
        end
        eta = params.rho * eta;
      end
      if not(sufficient)
        LogIf(params.debug && params.debug_decrease,...
              'Cancelling update of U(%d). No good stepsize found\n', i);
        unchanged = unchanged + 1;
      end
    else
      if params.no_adjusted_dependent_variates
        grad = alpha .* (w.* -err) * V + Ai;
        step = grad / covar;
        U(i,:) = U(i,:) - eta .* step;
      else
        adjx = alpha .* (uv + eta .* scaled_err);
        U(i,:) = (adjx * D * V + u*Gi - eta .* Ai) / covar;
      end
      stepsize(i) = eta;
    end
  end
end

change = mean_normchange(U, Uold);
step_report(factorname, change, unchanged, condition, m, stepsize, ...
            delta, params);
end % function newton_step_u

function [Z HZ delta change] = ...
    newton_step_z_dispatch(Y, V, Z, iter, HZ, params, glmvars)

if glmvars.alpha(2) == 0
  delta = 0;
  change = 0;
  return;
end

k = glmvars.k;
if params.add_bias_term
  error('not supported');
  pidx = [1:k k+3];
  Ii = diag(params.scale_i * [ones(1,k) 0]);
  [Z_tied HZ delta change] = ...
      newton_step_z(Y, V(:,pidx), Z(:,pidx), Ii, iter, HZ, 'Z', ...
                    params, glmvars);
  Z(:,pidx) = Z_tied;
else
  Ii = params.scale_i * eye(k);
  [Z HZ delta change] = ...
      newton_step_z(Y, V, Z, Ii, iter, HZ, 'Z', params, glmvars);
end

end % function newton_step_z_dispatch

function [Z, HZ, delta, change] = ...
    newton_step_z(Y, V, Z, Ii, iter, HZ, factorname, params, glmvars)

if glmvars.mock
  Z = normalize_rows(rand(size(Z)));
  delta = 0;
  change = NaN;
  return;
end

[n, r] = size(Y);
k = glmvars.k;
Zold = Z;
condition = repmat(NaN, 1, r);
stepsize = repmat(NaN, 1, r);
unchanged = 0;
delta = 0;
%[D didx] = create_scaling_matrix(V, params);

if strcmpi(factorname, 'Z')
  alpha = glmvars.alpha(2);
  wy_func = glmvars.wy_func;
  if params.inline_standard_weights
    pey = glmvars.wy_context.pey;
    disy = glmvars.wy_context.disy;
    Exclude = glmvars.wy_context.Exclude;
    Exclude_t = glmvars.wy_context.Exclude_trans;
  end
  dloss = @yloss;
  dloss_natural = @yloss_natural;
  inv_reg = params.inv_i;
elseif strcmpi(factorname, 'R')
  alpha = glmvars.alpha(3);
  wy_func = glmvars.wq_func;
  if params.inline_standard_weights
    pey = glmvars.wq_context.peq;
    disy = glmvars.wq_context.disq;
    Exclude = glmvars.wq_context.Exclude;
    Exclude_t = glmvars.wq_context.Exclude_trans;
  end
  dloss = @qloss;
  dloss_natural = @qloss_natural;
  inv_reg = params.inv_j;
else
  error('ack');
end

Vorig = V;
power = 1;
for i = 1:r
  power = TextProgressMessage(i, r, power, ...
                              sprintf('Updating %s:', factorname), ...
                              params.show_iter);

  if params.densify_rows
    y = full(Y(:,i));
  else
    y = Y(:,i);
  end
  z = Z(i,:);
  if glmvars.hasweights_y
    w = Wy(:,i);
  else
    row_idx = 1:rows(Vorig);
    col_idx = i * ones(1, rows(Vorig));
    if params.inline_standard_weights
      w = NaN * ones(size(y));
      idx = (y ~= 0);
      w(idx) = 1 / pey;
      w(~idx) = disy / pey;
      w(logical(Exclude(:,i))) = 0;
    else
      w = wy_func(y, row_idx, col_idx);
    end
  end
  [V s] = sample_factor_weighted(Vorig, w, params);
  y = y(s);
  w = w(s);
  vz = V*z';
  if all(w == 0)
    LogIf(params.debug && params.debug_zeroweights, ...
          sprintf('%s(%d): All weights w == 0\n', factorname), i);
  end
  expecty = params.ff(vz);
  err = (y - expecty)';
  raw_deriv = params.ff_prime(vz);
  deriv = replace_zero_deriv(raw_deriv, err', ...
                             sprintf('%s(%d)', factorname, i), params);
  scaled_err = err ./ deriv';
  %D(didx) = w .* raw_deriv;
  D = spdiags(w .* raw_deriv, 0, rows(V), rows(V));
  Ci = inv_reg(z);
  if params.add_bias_term
    Ci(k+1:end) = 0;
  end
  eta = initial_steplength(params, glmvars, iter);

  if glmvars.gradient
    Z(i,:) = Z(i,:) + eta * alpha * (w' .* err) * V - eta * Ci;
  elseif params.stochastic_opt
  grad = alpha .* (w' .* -err) * V + Ci;
  H_new = alpha .* V' * D * V + Ii;
  covar = estimate_stochastic_hessian(H_new, HZ{i}, iter);
  invertible = check_covariance(covar, params, factorname, i);
  if not(invertible)
    unchanged = unchanged + 1;
  else
    Z(i,:) = Z(i,:) - eta .* grad / covar;
    HZ{i} = covar;
    stepsize(i) = eta;
  end
  else
    covar = alpha .* V' * D * V + Ii;
    invertible = check_covariance(covar, params, factorname, i);
    if not(invertible)
      unchanged = unchanged + 1;
      continue;
    end
    if params.linesearch && iter >= params.min_armijo_iter
      grad = alpha .* -(w' .* err) * V + Ci;
      step = grad / covar;
      loss_old = alpha * dloss_natural(y, w, vz, V, z, params, glmvars);

      sufficient = false;
      while eta > params.mineta && not(sufficient)
        if params.no_adjusted_dependent_variates
          z_new = Z(i,:) - eta .* step;
        else
          adjx = alpha .* (vz' + eta .* scaled_err);
          z_new = (adjx * D * V + z*Ii - eta .* Ci) / covar;
        end
        loss_new = alpha * dloss(y, w, V, z_new, params, glmvars);
        sufficient = (loss_new - loss_old <= params.c * eta * step * grad');
        if sufficient
          Z(i,:) = z_new;
          stepsize(i) = eta;
          delta = delta + (loss_new - loss_old);
          break;
        end
        eta = params.rho * eta;
      end
      if not(sufficient)
        LogIf(params.debug && params.debug_decrease,...
              'Cancelling update of %s(%d). No good stepsize found\n', ...
              factorname, i);
        unchanged = unchanged + 1;
      end
    else
      if params.no_adjusted_dependent_variates
        grad = alpha .* -(w' .* err) * V + Ci;
        step = grad / covar;
        Z(i,:) = Z(i,:) - eta .* step;
      else
        adjx = alpha .* (vz' + eta .* scaled_err);
        Z(i,:) = (adjx * D * V + z*Ii - eta .* Ci) / covar;
      end
      stepsize(i) = eta;
    end
  end
end

change = mean_normchange(Z, Zold);
step_report(factorname, change, unchanged, condition, r, stepsize, ...
            delta, params);
end % newton_step_z

function [R HR delta change] = ...
    newton_step_r_dispatch(Q, V, R, iter, HR, params, glmvars)

if glmvars.alpha(3) == 0
  delta = 0;
  change = 0;
  return;
end

k = glmvars.k;
if params.add_bias_term
  error('not supported');
  pidx = [1:k k+3];
  Ji = diag(params.scale_j * [ones(1,k) 0]);
  [R_tied HR delta change] = ...
      newton_step_z(Q, V(:,pidx), R(:,pidx), Ji, ...
                    iter, HR, 'R', @qloss, @qloss_natural, ...
                    params, glmvars);
  R(:,pidx) = R_tied;
else
  Ji = params.scale_j * eye(k);
  [R HR delta change] = ...
      newton_step_z(Q, V, R, Ji, iter, HR, 'R', params, glmvars);
end

end % function newton_step_z_dispatch

function [V, HV, delta, change] = newton_step_v_dispatch(X, Yt, Qt, U, V, Z, R, iter, HV, params, glmvars)

k = glmvars.k;
alpha = glmvars.alpha;

if params.add_bias_term
  error('broken');
  pidx = [1:k k+1 k+4];
  Hi = diag(params.scale_h * [ones(1,k) 0 0]);
  [V_tied HV delta change] = ...
      newton_step_v(X, Yt, U(:,pidx), V(:,pidx), Z(:,pidx), Hi, ...
                    iter, HV, params, glmvars);
  V(:,pidx) = V_tied;
else
  Hi = params.scale_h * eye(k);
  [V HV delta change] = newton_step_v(X,Yt,Qt,U,V,Z,R,Hi,iter,HV,params,glmvars);
end

end % newton_step_v_dispatch

function [V, HV, delta, change] = newton_step_v(X, Yt, Qt, U, V, Z, R, Hi, iter, HV, params, ...
                                            glmvars)
if glmvars.mock
  V = normalize_rows(rand(size(V)));
  delta = 0;
  change = NaN;
  return;
end

[m, n] = size(X);
[r, q] = size(Yt);
[s, c] = size(Qt);
k = glmvars.k;
factorname = 'V';
Vold = V;

alpha = glmvars.alpha;
condition = repmat(NaN, 1, n);
stepsize = repmat(NaN, 1, n);
unchanged = 0;
delta = 0;
% $$$ [Dx didx_x] = create_scaling_matrix(U, params);
% $$$ [Dy didx_y] = create_scaling_matrix(Z, params);
% $$$ [Dq didx_q] = create_scaling_matrix(R, params);
wx_func = glmvars.wx_func;
wy_func = glmvars.wy_func;
wq_func = glmvars.wq_func;

if params.inline_standard_weights
  pex = glmvars.wx_context.pex;
  disx = glmvars.wx_context.disx;
  pey = glmvars.wy_context.pey;
  disy = glmvars.wy_context.disy;
  peq = glmvars.wq_context.peq;
  disq = glmvars.wq_context.disq;
  Exclude_wx = glmvars.wx_context.Exclude;
  Exclude_wx_t = glmvars.wx_context.Exclude_trans;
  Exclude_wy = glmvars.wy_context.Exclude;
  Exclude_wy_t = glmvars.wy_context.Exclude_trans;
  Exclude_wq = glmvars.wq_context.Exclude;
  Exclude_wq_t = glmvars.wq_context.Exclude_trans;
end

Uorig = U;
Zorig = Z;
Rorig = R;
power = 1;
for i = 1:n
  power = TextProgressMessage(i, n, power, 'Updating V: ', params.show_iter);

  v = V(i,:);
  if params.densify_rows
    x = full(X(:,i));
  else
    x = X(:,i);
  end
  if glmvars.hasweights_x
    wx = Wx(:,i);
  else
    row_idx = 1:rows(Uorig);
    col_idx = i * ones(1, rows(Uorig));
    if params.inline_standard_weights
      wx = NaN * ones(size(x));
      idx = (x ~= 0);
      wx(idx) = 1 / pex;
      wx(~idx) = disx / pex;
      wx(logical(Exclude_wx(:,i))) = 0;
    else
      wx = wx_func(x, row_idx, col_idx);
    end
  end
  [U s] = sample_factor_weighted(Uorig, wx, params);
  x = x(s);
  wx = wx(s);
  uv = U*v';
  expectx = params.f(uv);
  errx = (x - expectx)';
  raw_derivx = params.f_prime(uv);
  derivx = replace_zero_deriv(raw_derivx, errx', sprintf('V(%d)', i), params);
  scaled_errx = errx ./ derivx';
  %Dx(didx_x) = wx .* raw_derivx;
  Dx = spdiags(wx .* raw_derivx, 0, rows(U), rows(U));

  if params.densify_rows
    y = full(Yt(:,i))';
  else
    y = Yt(:,i)';
  end
  if glmvars.hasweights_y
    wy = Wy(i,:);
  else
    row_idx = i * ones(1, rows(Zorig));
    col_idx = 1:rows(Zorig);
    if params.inline_standard_weights
      wy = NaN * ones(size(y));
      idx = (y ~= 0);
      wy(idx) = 1 / pey;
      wy(~idx) = disy / pey;
      wy(logical(Exclude_wy_t(:,i))) = 0;
    else
      wy = wy_func(y, row_idx, col_idx);
    end
  end
  [Z t] = sample_factor_weighted(Zorig, wy, params);
  y = y(t);
  wy = wy(t);
  vz = v*Z';
  expecty = params.ff(vz);
  erry = (y - expecty);
  raw_derivy = params.ff_prime(vz);
  derivy = replace_zero_deriv(raw_derivy, erry, sprintf('V(%d)', i), params);
  scaled_erry = erry ./ derivy;
  %Dy(didx_y) = wy .* raw_derivy;
  Dy = spdiags((wy .* raw_derivy)', 0, rows(Z), rows(Z));

  if params.densify_rows
    q = full(Qt(:,i))';
  else
    q = Qt(:,i)';
  end
  if glmvars.hasweights_q
    wq = Wq(i,:);
  else
    row_idx = i * ones(1, rows(Rorig));
    col_idx = 1:rows(Rorig);
    if params.inline_standard_weights
      wq = NaN * ones(size(q));
      idx = (q ~= 0);
      wq(idx) = 1/peq;
      wq(~idx) = disq / peq;
      wq(logical(Exclude_wq_t(:,i))) = 0;
    else
      wq = wq_func(q, row_idx, col_idx);
    end
  end
  [R r] = sample_factor_weighted(Rorig, wq, params);
  q = q(r);
  wq = wq(r);
  if all(wy == 0) && all(wx == 0) && all(wq == 0)
    LogIf(params.debug && params.debug_zeroweights, ...
          'V(%d): All weights wy == 0\n', i);
  end
  vr = v*R';
  expectq = params.fff(vr);
  errq = (q - expectq);
  raw_derivq = params.fff_prime(vr);
  derivq = replace_zero_deriv(raw_derivq, errq, sprintf('V(%d)', i), params);
  scaled_errq = errq ./ derivq;
  %Dq(didx_q) = wq .* raw_derivq;
  Dq = spdiags((wq .* raw_derivq)', 0, rows(R), rows(R));

  Bi = params.inv_i(v);
  if params.add_bias_term
    Bi(k+1:end) = 0;
  end
  eta = initial_steplength(params, glmvars, iter);

  if glmvars.gradient
    grad = alpha(1) .* (wx' .* -errx) * U + ...
           alpha(2) .* (wy .* -erry) * Z + ...
           alpha(3) .* (wq .* -errq) * R + Bi;
    V(i,:) = V(i,:) - eta .* grad;
  elseif params.stochastic_opt
    grad = alpha(1) .* (wx' .* -errx) * U + ...
           alpha(2) .* (wy .* -erry) * Z + ...
           alpha(3) .* (wq .* -errq) * R + Bi;
    H_new = alpha(1) * (U' * Dx * U) + ...
            alpha(2) * (Z' * Dy * Z) + ...
            alpha(3) * (R' * Dq * R) + Hi;
    covar = estimate_stochastic_hessian(H_new, HV{i}, iter);
    invertible = check_covariance(covar, params, factorname, i);
    if not(invertible)
      unchanged = unchanged + 1;
    else
      V(i,:) = V(i,:) - eta .* grad / covar;
      HV{i} = covar;
      stepsize(i) = eta;
    end
  else
    covar = alpha(1) * (U' * Dx * U) + ...
            alpha(2) * (Z' * Dy * Z) + ...
            alpha(3) * (R' * Dq * R) + Hi;
    invertible = check_covariance(covar, params, factorname, i);
    if not(invertible)
      unchanged = unchanged + 1;
      continue;
    end
    if params.linesearch && iter >= params.min_armijo_iter
      grad = alpha(1) .* (wx' .* -errx) * U + ...
             alpha(2) .* (wy .* -erry) * Z + ...
             alpha(3) .* (wq .* -errq) * R + Bi;
      step = grad / covar;
      t1 = xloss_natural(x, wx, uv, U, v, params, glmvars);
      t2 = yloss_natural(y, wy, vz, v, Z, params, glmvars);
      t3 = qloss_natural(q, wq, vr, v, R, params, glmvars);
      loss_old = alpha(1) * t1 + alpha(2) * t2 + alpha(3) * t3;

      sufficient = false;
      while eta > params.mineta && not(sufficient)
        if params.no_adjusted_dependent_variates
          v_new = V(i,:) - eta .* step;
        else
          adjx = alpha(1) .* (uv' + eta .* scaled_errx);
          adjy = alpha(2) .* (vz + eta .* scaled_erry);
          adjq = alpha(3) .* (vr + eta .* scaled_errq);
          adjusted = (adjx * Dx * U) + (adjy * Dy * Z) + (adjq * Dq * R) + ...
              v*Hi - eta.*Bi;
          v_new = adjusted / covar;
          %v_new = adjusted * pinv(covar);
        end
        t1 = xloss(x, wx, U, v_new, params, glmvars);
        t2 = yloss(y, wy, v_new, Z, params, glmvars);
        t3 = qloss(q, wq, v_new, R, params, glmvars);
        loss_new = alpha(1) * t1 + alpha(2) * t2 + alpha(3) * t3;
        sufficient = (loss_new - loss_old <= params.c * eta * step * grad');
        if sufficient
          V(i,:) = v_new(:);
          stepsize(i) = eta;
          delta = delta + (loss_new - loss_old);
          break;
        end
        eta = params.rho * eta;
      end
      if not(sufficient)
        LogIf(params.debug && params.debug_decrease,...
              'Cancelling update of V(%d). No good stepsize found.\n', i);
        unchanged = unchanged + 1;
      end
    else
      if params.no_adjusted_dependent_variates
        grad = alpha(1) .* (wx' .* -errx) * U + ...
               alpha(2) .* (wy .* -erry) * Z + ...
               alpha(3) .* (wq .* -errq) * R + Bi;
        step = grad / covar;
        V(i,:) = V(i,:) - eta .* step;
      else
        adjx = alpha(1) .* (uv' + eta .* scaled_errx);
        adjy = alpha(2) .* (vz + eta .* scaled_erry);
        adjq = alpha(3) .* (vr + eta .* scaled_errq);
        adjusted = (adjx * Dx * U) + (adjy * Dy * Z) + (adjq * Dq * R) + ...
            v*Hi - eta.*Bi;
        v_new = adjusted / covar;
        V(i,:) = v_new;
      end
      stepsize(i) = eta;
    end
  end
end

change = mean_normchange(V, Vold);
step_report(factorname, change, unchanged, condition, n, stepsize, ...
            delta, params);
end % newton_step_v

function loss = xloss_natural_noreg(X,W,T,params)

loss = full(params.loss_f(X, W, T, params.f));
end % function xloss_natural_noreg

function obj = xloss_natural(X,W,T,U,V,params,glmvars)
% xloss_natural -- Compute the reconstruction loss of the X matrix, given
% the factors and the natural parameters.
%
% obj = xloss_natural(X,W,T,U,V,params)
% Computes the reconstruction error of X by f(T). The data, weights, and
% parameters can be either vectors or matrices.
%
% Arguments:
%
% X: Data matrix (m x n)
% W: Weight matrix (m x n)
% T: Natural parameters (m x n)
% U: Row factors (m x k)
% V: Column factors (n x k)
% params: Optimization parameters
% glmvars: Variables scoped throughout glm.m
%
% Return Values:
%
% obj: Reconstruction error

loss = full(params.loss_f(X, W, T, params.f));

k = glmvars.k;
if cols(U) > k
  U = U(:,1:k);
end
if cols(V) > k
  V = V(:,1:k);
end
reg_u = params.scale_g * sum(U(:).^2)/(2*numel(U));
reg_v = params.scale_h * sum(V(:).^2)/(2*3*numel(V));
obj = loss + reg_u + reg_v;
%fprintf(1, 'X: loss = %g, reg_u = %g, reg_v = %g\n', loss, reg_u, reg_v);

end % function xloss_natural

function loss = yloss_natural_noreg(Y,W,T,params)

loss = full(params.loss_ff(Y, W, T, params.ff));
end % function yloss_natural_noreg

function obj = yloss_natural(Y,W,T,V,Z,params,glmvars)
% xloss_natural -- Compute the reconstruction loss of the Y matrix, given
% the factors and the natural parameters.
%
% obj = yloss_natural(Y,W,T,V,Z,params)
% Computes the reconstruction error of Y by f(T). The data, weights, and
% parameters can be either vectors or matrices.
%
% Arguments:
%
% Y: Data matrix (n x r)
% W: Weight matrix (n x r)
% T: Natural parameters (n x r)
% V: Row factors (n x k)
% Z: Column factors (r x k)
% params: Optimization parameters
%
% Return Values:
%
% obj: Reconstruction error

loss = full(params.loss_ff(Y, W, T, params.ff));

k = glmvars.k;
if cols(V) > k
  V = V(:,1:k);
end
if cols(Z) > k
  Z = Z(:,1:k);
end
reg_v = params.scale_h * sum(V(:).^2)/(2*3*numel(V));
reg_z = params.scale_i * sum(Z(:).^2)/(2*numel(Z));
obj = loss + reg_v + reg_z;
%fprintf(1, 'Y: loss = %g, reg_v = %g, reg_z = %g\n', loss, reg_v, reg_z);

end % function xloss_natural

function loss = qloss_natural_noreg(Q,W,T,params)

loss = full(params.loss_fff(Q, W, T, params.fff));
end % function qloss_natural_noreg

function loss = qloss_natural(Q, W, T, V, R, params, glmvars)

loss = full(params.loss_fff(Q, W, T, params.fff));

k = glmvars.k;
if cols(V) > k
  V = V(:,1:k);
end
if cols(R) > k
  R = R(:,1:k);
end

reg_v = params.scale_h * sum(sum(V.^2))/(2*3*numel(V));
reg_r = params.scale_j * sum(sum(R.^2))/(2*numel(R));
obj = loss + reg_v + reg_r;
end

function obj = xloss(X,W,U,V,params,glmvars)
% loss_x -- Compute the reconstruction loss of the X matrix.
%
% obj = loss_x(X,W,U,V,params)
% Computes the weighted reconstruction error of X as f(UV') where
% U = g(A) and V = h(B). The functions works regardless of whether
% the data and the reconstructions are vectors or matrices.
%
% When params.alpha == 0, this function always returns 0.
%
% Arguments:
%
% X: Matrix we want to approximate
% W: Reconstruction weights. Precondition: size(X) == size(W)
% U: Row factors for X. Precondition: rows(X) == rows(U)
% V: Col factors for X. Precondition: cols(X) == rows(V)
% params: Optimization parameters. Precondition: isfield(params,'alpha'),
%         isfield(params,'scale_g'), isfield(params,'scale_h'),
%         isfield(params,'f')
%
% Return Values:
%
% obj: Reconstruction error D_F(UV'|X,W) + D_G(0|U) + D_H(0|V). Note,
%      the error has not been scaled by alpha.

if rows(X) ~= rows(U)
  error('RFA:InvalidArgument', 'rows(X) ~= rows(U) (%d vs. %d)',...
        rows(X), rows(U));
end
if cols(X) ~= rows(V)
  error('RFA:InvalidArgument', 'cols(X) ~= rows(V) (%d vs. %d)',...
        cols(X), rows(V));
end
if any(size(X) ~= size(W))
  error('RFA:InvalidArgument',...
        'Data and weight matrices do not conform (%d x %d vs. %d x %d)',...
        rows(X), cols(X), rows(W), cols(W));
end

check_dims(X, rows(W), cols(W), 'X');
if glmvars.alpha(1) ~= 0
  obj = xloss_natural(X, W, U*V', U, V, params, glmvars);
else
  obj = 0;
end

end % function xloss

function obj = yloss(Y,W,V,Z,params,glmvars)
% yloss -- Compute the reconstruction loss of the Y matrix
%
% obj = yloss(Y,W,V,Z,params)
% Computes the reconstruction error of Y approximated by ff(VZ') where
% V = h(B) and Z = i(C).
%
% If alpha == 1 then this function always returns 0.
%
% Arguments:
%
% Y: Matrix to reconstruct.
% W: Reconstruction weights. Precondition: size(Y) == size(W).
% V: Row factors for Y. Precondition: rows(Y) == rows(V).
% Z: Col factors for Y. Precondition: cols(Y) == rows(Z).
% params: Optimization parameters. Precondition: isfield(params,'alpha'),
%         isfield(params,'scale_h'), isfield(params,'scale_i'),
%         isfield(params,'ff').
%
% Return Values:
%
% obj: Reconstruction error.

if rows(Y) ~= rows(V)
  error('RFA:InvalidArgument', 'rows(Y) ~= rows(V) (%d vs. %d)',...
        rows(Y), rows(V));
end
if cols(Y) ~= rows(Z)
  error('RFA:InvalidArgument', 'cols(X) ~= rows(V) (%d vs. %d)',...
        cols(Y), rows(Z));
end
if any(size(Y) ~= size(W))
  error('RFA:InvalidArgument',...
        'Data and weight matrices do not conform (%d x %d vs. %d x %d)',...
        rows(Y), cols(Y), rows(W), cols(W));
end

if glmvars.alpha(2) ~= 0
  obj = yloss_natural(Y, W, V*Z', V, Z, params, glmvars);
else
  obj = 0;
end

end % function

function obj = qloss(Q, W, V, R, params, glmvars)

if rows(Q) ~= rows(V)
  error('RFA:InvalidArgument', 'rows(X) ~= rows(U) (%d vs. %d)',...
        rows(X), rows(U));
end
if cols(Q) ~= rows(R)
  error('RFA:InvalidArgument', 'cols(X) ~= rows(V) (%d vs. %d)',...
        cols(X), rows(V));
end
if any(size(Q) ~= size(W))
  error('RFA:InvalidArgument',...
        'Data and weight matrices do not conform (%d x %d vs. %d x %d)',...
        rows(X), cols(X), rows(W), cols(W));
end

if glmvars.alpha(3) ~= 0
  obj = qloss_natural(Q, W, V*R', V, R, params, glmvars);
else
  obj = 0;
end

end % function

function obj = loss(Xt, Y, Qt, U, V, Z, R, params, glmvars)
% loss -- Compute the overall loss of the three entity-type model.
%
% obj = loss(X,Y,U,V,Z,params)
% Compute the overall loss of the three entity-type model:
%    alpha * [D_F(UV'|X) + D_G(0|U) + D_H(0|V)] +
%    (1-alpha) * [D_F(VZ'|Y) + D_H(0|V) + D_I(0|Z)]
%
% Implementation Details:
%
% Designed to compute the loss of each row of X and each column of
% Y, which precludes creating a large dense weight matrix.
%
% Arguments:
%
% Xt: An n x m matrix.
% Y: An n x r matrix.
% U: An m x k matrix of factors.
% V: An n x k matrix of shared factors.
% Z: An r x k matrix of factors.
% params: Optimization parameters. Precondition: isfield(params,'alpha')
% glmvars: Variables scoped throughout glm.m
%
% Return Values:
%
% obj: Reconstruction loss.

if glmvars.mock || glmvars.mock_loss
  obj = rand;
  return;
end

Vorig = V;
k = glmvars.k;

obj = 0;
alpha = glmvars.alpha;
m = cols(Xt);
n = rows(Xt);
r = cols(Y);
s = rows(Qt);
wx_func = glmvars.wx_func;
wy_func = glmvars.wy_func;
wq_func = glmvars.wq_func;

pex = glmvars.wx_context.pex;
disx = glmvars.wx_context.disx;
pey = glmvars.wy_context.pey;
disy = glmvars.wy_context.disy;
peq = glmvars.wq_context.peq;
disq = glmvars.wq_context.disq;

Exclude_t = glmvars.wx_context.Exclude_trans;
row_idx = repmat(NaN, 1, n);
column_idx = 1:n;
for i = 1:m
  if params.densify_rows
    x = full(Xt(:,i))';
  else
    x = Xt(:,i)';
  end
  if glmvars.hasweights_x
    w = glmvars.Wx(i,:);
  else
    row_idx(:) = i;
    if params.inline_standard_weights
      w = NaN * ones(size(x));
      idx = (x ~= 0);
      w(idx) = 1 / pex;
      w(~idx) = disx / pex;
      w(logical(Exclude_t(:,i))) = 0;
    else
      w = wx_func(x, row_idx, column_idx);
    end
  end
  u = U(i,:);
  obj = obj + alpha(1) * xloss_natural_noreg(x, w, u*V', params);
end

Exclude = glmvars.wy_context.Exclude;
row_idx = 1:n;
column_idx = repmat(NaN, 1, n);
for i = 1:r
  if params.densify_rows
    y = full(Y(:,i));
  else
    y = Y(:,i);
  end
  if glmvars.hasweights_y
    w = glmvars.Wy(:,i);
  else
    column_idx(:) = i;
    if params.inline_standard_weights
      w = NaN * ones(size(y));
      idx = (y ~= 0);
      w(idx) = 1 / pey;
      w(~idx) = disy / pey;
      w(logical(Exclude(:,i))) = 0;
    else
      w = wy_func(y, row_idx, column_idx);
    end
  end
  z = Z(i,:);
  obj = obj + alpha(2) * yloss_natural_noreg(y, w, V*z', params);
end

Exclude_t = glmvars.wq_context.Exclude_trans;
row_idx = repmat(NaN, 1, s);
column_idx = 1:s;
for i = 1:n
  if params.densify_rows
    q = full(Qt(:,i))';
  else
    q = Qt(:,i)';
  end
  if glmvars.hasweights_q
    w = glmvars.Wq(i,:)
  else
    row_idx(:) = i;
    if params.inline_standard_weights
      w = NaN * ones(size(q));
      idx = (q ~= 0);
      w(idx) = 1/peq;
      w(~idx) = disq / peq;
      w(logical(Exclude_t(:,i))) = 0;
    else
      w = glmvars.wq_func(q, row_idx, column_idx);
    end
  end
  v = V(i,:);
  obj = obj + alpha(3) * qloss_natural_noreg(q, w, v*R', params);
end

U = U(:,1:k);
V = V(:,1:k);
Z = Z(:,1:k);
R = R(:,1:k);
obj = obj + params.scale_g * sum(U(:).^2)/(2*numel(U));
obj = obj + params.scale_h * sum(V(:).^2)/(2*numel(V));
obj = obj + params.scale_i * sum(Z(:).^2)/(2*numel(Z));
end % function
