function [U V obj debug] = mmmf(X, params, varargin)
% MMMF - Max-Margin Matrix Factorization
%
% [U V obj debug] = mmmf(X, params, varargin)
% Compute a low rank representation of a matrix X with ordinal entries 1..R
% and a binary loss function L_i -- e.g., logistic or shift-generalized logistic
%
% Loss Function: L(U,V) = \sum_{i=1}^R L_i(U,V) + Regularizer(U, V),
% Binarization: X = (X < i)
% Regularizer: Squared Loss
%
% Arguments:
%
% X: Data matrix (m x n, zero entries assumed to be missing)
% params: Optimization and model parameters (see @glmparams)
% varargin: Optional parameters (not currently used).
%
% Return Values:
%
% U: Factor matrix (m x k, k = params.dims)
% V: Factor matrix (m x k)
% obj: The objective at each iteration
% debug: Extra debugging information (not currently used).

% Author: Ajit Singh (ajit@cs.cmu.edu)

error(nargchk(2, Inf, nargin));

[strict_mmmf gradient clustering unused] = process_options(varargin, 'strict_mmmf', true, ...
                                         'gradient', false, 'clustering', false);

check_matrix(X);

[m, n] = size(X);
Xt = X';

params = struct(params);
k = params.dims;
glmvars.k = k;
values = sort(full(setdiff(unique(X), 0)));
values = values(2:end); % instead of {<1,<2,<3,<4,<5} thresholds are
                        % {<2, <3, <4, <5} so we match #params in Rennie code.
b = length(values);
glmvars.b = b;
glmvars.values = values;
glmvars.gradient = gradient;
glmvars.strict_mmmf = strict_mmmf;
glmvars.clustering = clustering;

WarnIf(glmvars.gradient, 'Using gradient method');
WarnIf(not(glmvars.strict_mmmf), 'Using MMMF with column and row biases');
assert(not(glmvars.clustering), 'Cannot cluster with bias terms !');

glmvars.numel = nnz(X);

U = [normalize_rows(rand(m, k)) rand(m, b)];
if glmvars.strict_mmmf
  V = [normalize_rows(rand(n, k)) ones(n, b)];
else
  V = [normalize_rows(rand(n, k)) rand(n, b)];
end
debug.U = cell(1, params.maxiter);
debug.V = cell(1, params.maxiter);
debug.U{1} = U;
debug.V{2} = V;

HU = cell(1, m);
HV = cell(1, n);
cU = repmat(false, 1, m);
cV = repmat(false, 1, n);

debug.cputime = ( 0 );
obj = ( loss(Xt, U, V, params, glmvars) );
fprintf(1, 'Iteration 0 (random parameters): obj = %g\n', obj(1));

for iter = 1:params.maxiter
  start_time = cputime;
  [Un HU cU change_u] = newton_step_u(Xt, U, V, iter, HU, cU, params, glmvars);
  [Vn HV cV change_v] = newton_step_v(X, Un, V, iter, HV, cV, params, glmvars);
  duration = cputime - start_time;
  epsilon = loss(Xt, Un, Vn, params, glmvars) - obj(end);
  if isfinite(epsilon) && (epsilon <= 0 || params.stochastic_opt)
    obj = [obj obj(end) + epsilon];
    U = Un;
    V = Vn;
    debug.U{iter} = U;
    debug.V{iter} = V;
    debug.cputime = [ debug.cputime debug.cputime(end) + duration ];
  elseif not(isfinite(epsilon))
    if params.stochastic_opt
      warning('RFA:BadResult', 'Infinite loss...throwing away the update');
    else
      error('RFA:BadResult', 'Infinite loss on non-stochastic descent');
    end
  else
    warning('RFA:BadResult', 'Freaky, epsilon = %e\n', epsilon);
    if not(params.stochastic_opt)
      break;
    end
  end

  oc = abs(epsilon/obj(end));
  if mod(iter, 1) == 0
    fprintf(1, ['Iteration %d: eps = %g, obj = %g, '...
                'nc(u/v) = %.1e/%.1e, oc = %.3f, time = %.3f sec\n'], ...
            iter, epsilon, obj(end), change_u, change_v, oc, duration);
  end
  if not(params.stochastic_opt)
    if oc < params.obj_threshold && iter >= params.miniter
      fprintf(1, 'oc below threshold (%f)\n', oc);
      break;
    end
  end
end
end

function [U HU quit change] = newton_step_u(Xt, U, V, iter, HU, quit, params, glmvars)

[n, m] = size(Xt);

k = glmvars.k;
Uold = U;
factorname = 'U';
condition = repmat(NaN, 1, m);
stepsize = repmat(NaN, 1, m);
delta = 0;
unchanged = sum(quit);
Gi = params.scale_g * eye(cols(U));
Gi(k+1:end, k+1:end) = 0;

function [p loss_new] = update_u(x, w, u, v, step, eta, params, glmvars)
p = u - eta .* step;
u_new = p;
loss_new = NaN;
if nargout > 1
  loss_new = xloss(x, w, u_new, v, params, glmvars);
end
end

Vorig = V;
power = 1;

Hessian = step_u_hessian(x, w, u, V, Gi, params, glmvars);
[R p] = chol(Hessian);
if p ~= 0
  Hessian = eye(rows(Hessian));
  R = eye(rows(Hessian));
end
Ri = inv(R);
inv_covar = Ri * Ri';
covar = Hessian;

for i = find(~quit)
  power = TextProgressMessage(i, m, power, 'Updating U: ', params.show_iter);

  u = U(i,:);
  x = Xt(:,i)';
  w = zeros(size(x));
  w(x ~= 0) = 1/glmvars.numel;
  [V s] = sample_factor_weighted(Vorig, w, params);
  w = w(s);
  x = x(s);
  if rows(V) == 0
    unchanged = unchanged + 1;
    continue;
  end

  grad = step_u_gradient(x, w, u, V, params, glmvars);

  eta = initial_steplength(params, glmvars, iter);
  if params.stochastic_opt
    % Must recompute the Hessian here
    error('Broken');
    covar = estimate_stochastic_hessian(Hessian, HU{i}, iter);
    step = step_direction(grad, covar, inv_covar, params, glmvars);
    U(i,:) = U(i,:) - eta .* step;
    HU{i} = covar;
    stepsize(i) = eta;
  else
    [step die_now] = step_direction(grad, covar, inv_covar, params, glmvars);
    if die_now
      unchanged = unchanged + 1;
      continue;
    end
    % TODO(ajit): Change linesearch to update, and in update check whether
    % or not we are required to do the linesearch.
    f = @(eta)(update_u(x, w, u, V, step, eta, params, glmvars));
    armijo = @(new, old, eta)(new - old <= params.c * eta * step * grad');
    [U(i,:) stepsize(i) change] = linesearch(f, u, eta, armijo, i, params);
    delta = delta + change;
    if not(isfinite(stepsize(i)))
      LogIf(params.debug && params.debug_decrease, ...
            'Cancelling update of %s(%d). No good stepsize found\n', ...
            factorname, i);
      unchanged = unchanged + 1;
    end
  end
end

change = mean_normchange(U, Uold);
step_report(factorname, change, unchanged, condition, m, stepsize, ...
            delta, params);
end

function grad = step_u_gradient(x_ordinal, w, uorig, Vorig, params, glmvars)

k = glmvars.k;
b = glmvars.b;
assert(cols(uorig) == k + b);
assert(cols(Vorig) == k + b);

if rows(Vorig) == 0
  grad = zeros(1, k+b);
  return;
end

values = glmvars.values;

Ai = params.inv_g(uorig);
Ai(k+1:end) = 0;
grad = Ai;

for r = 1:b
  latents = [1:k k+r];
  x = x_ordinal;
  x(x < values(r)) = 1;
  x(x >= values(r)) = 0;

  % Make predictions using all the latent factors and compute residuals.
  uv = uorig(:,latents) * Vorig(:,latents)';
  expectx = params.f(uv);
  err = (x - expectx);

  % Add in the contribution of the gradient.
  V = Vorig(:,latents);
  grad(latents) = grad(latents) + (w .* -err) * V;
end
end

function covar = step_u_hessian(x_ordinal, w, uorig, Vorig, Gi, params, glmvars)

k = glmvars.k;
b = glmvars.b;
assert(cols(uorig) == k + b);
assert(cols(Vorig) == k + b);

if rows(Vorig) == 0
  grad = zeros(1, k+b);
  H = eye(k+b);
  return;
end

values = glmvars.values;

covar = Gi;
D = eye(rows(Vorig));
didx = find(D);

for r = 1:b
  latents = [1:k k+r];
  x = x_ordinal;
  x(x < values(r)) = 1;
  x(x >= values(r)) = 0;

  % Compute the coordinate scalings for the Hessian -- i.e., the weights
  % in iteratively reweighted least squares.
  uv = uorig(:,latents) * Vorig(:,latents)';
  raw_deriv = params.f_prime(uv);
  %D = spdiags((w .* raw_deriv)', 0, rows(Vorig), rows(Vorig));
  D(didx) = w .* raw_deriv;

  % Add in the contribution of the Hessian and the gradient.
  V = Vorig(:,latents);
  covar(latents, latents) = covar(latents, latents) + V' * D * V;
end
end

% $$$ function [grad H] = step_u_grad_hessian(x_ordinal, w, uorig, Vorig, Gi, params, glmvars)
% $$$ 
% $$$ k = glmvars.k;
% $$$ b = glmvars.b;
% $$$ assert(cols(uorig) == k + b);
% $$$ assert(cols(Vorig) == k + b);
% $$$ 
% $$$ if rows(Vorig) == 0
% $$$   grad = zeros(1, k+b);
% $$$   H = eye(k+b);
% $$$   return;
% $$$ end
% $$$ 
% $$$ values = glmvars.values;
% $$$ 
% $$$ Ai = params.inv_g(uorig);
% $$$ Ai(k+1:end) = 0;
% $$$ 
% $$$ grad = Ai;
% $$$ H = Gi;
% $$$ 
% $$$ D = eye(rows(Vorig));
% $$$ didx = find(D);
% $$$ 
% $$$ for r = 1:b
% $$$   latents = [1:k k+r];
% $$$   x = x_ordinal;
% $$$   x(x < values(r)) = 1;
% $$$   x(x >= values(r)) = 0;
% $$$ 
% $$$   % Make predictions using all the latent factors and compute residuals.
% $$$   uv = uorig(:,latents) * Vorig(:,latents)';
% $$$   expectx = params.f(uv);
% $$$   err = (x - expectx);
% $$$   raw_deriv = params.f_prime(uv);
% $$$   %D = spdiags((w .* raw_deriv)', 0, rows(Vorig), rows(Vorig));
% $$$   D(didx) = w .* raw_deriv;
% $$$ 
% $$$   % Add in the contribution of the Hessian and the gradient.
% $$$   V = Vorig(:,latents);
% $$$   grad(latents) = grad(latents) + (w .* -err) * V;
% $$$   H(latents, latents) = H(latents, latents) + V' * D * V;
% $$$ end
% $$$ end

function [V HV quit change] = newton_step_v(X, U, V, iter, HV, quit, params, glmvars)

[m, n] = size(X);
k = glmvars.k;
factorname = 'V';
Vold = V;
condition = repmat(NaN, 1, n);
stepsize = repmat(NaN, 1, n);
delta = 0;
unchanged = sum(quit);
Hi = params.scale_h * eye(cols(V));
Hi(k+1:end,k+1:end) = 0;

function [p loss_new] = update_v(x, w, u, v, step, eta, params, glmvars)
p = v - eta .* step;
if glmvars.strict_mmmf
  p = [p(1:k) v(k+1:end)];
end
v_new = p;
loss_new = NaN;
if nargout > 1
  loss_new = xloss(x, w, u, v_new, params, glmvars);
end
end

Uorig = U;
power = 1;

for i = find(~quit)
  power = TextProgressMessage(i, n, power, 'Updating V: ', params.show_iter);

  v = V(i,:);
  x = X(:,i);
  w = zeros(size(x));
  w(x ~= 0) = 1/glmvars.numel;
  [U s] = sample_factor_weighted(Uorig, w, params);
  w = w(s);
  x = x(s);
  if rows(U) == 0
    unchanged = unchanged + 1;
    continue;
  end

  [grad Hessian] = step_v_grad_hessian(x, w, U, v, Hi, params, glmvars);

  eta = initial_steplength(params, glmvars, iter);
  if params.stochastic_opt
    covar = estimate_stochastic_hessian(Hessian, HV{i}, iter);
    step = step_direction(grad, covar, params, glmvars);
    V(i,:) = V(i,:) - eta .* step;
    HV{i} = covar;
    stepsize(i) = eta;
  else
    [step die_now] = step_direction(grad, Hessian, params, glmvars);
    if die_now
      unchanged = unchanged + 1;
      continue;
    end
    f = @(eta)(update_v(x, w, U, v, step, eta, params, glmvars));
    armijo = @(new, old, eta)(new - old <= params.c * eta * step * grad');
    [V(i,:) stepsize(i) change] = linesearch(f, v, eta, armijo, i, params);
    delta = delta + change;
    if not(isfinite(stepsize(i)))
      LogIf(params.debug && params.debug_decrease, ...
            'Cancelling update of %s(%d). No good stepsize found\n', ...
            factorname, i);
      unchanged = unchanged + 1;
    end
  end
end

change = mean_normchange(V, Vold);
step_report(factorname, change, unchanged, condition, n, stepsize, ...
            delta, params);
end

function grad = step_v_grad(x_ordinal, w, Uorig, vorig, params, glmvars)

k = glmvars.k;
b = glmvars.b;
assert(cols(vorig) == k + b);
assert(cols(Uorig) == k + b);

if rows(Uorig) == 0
  grad = zeros(1,k+b);
  return;
end

values = glmvars.values;

Bi = params.inv_h(vorig);
Bi(k+1:end) = 0;
grad = Bi;

for r = 1:b
  latents = [1:k k+r];
  x = x_ordinal;
  x(x < values(r)) = 1;
  x(x >= values(r)) = 0;

  % Make predictions using all the latent factors and compute residuals
  uv = Uorig(:,latents) * vorig(:,latents)';
  expectx = params.f(uv);
  err = (x - expectx)';

  % Add in the contribution of the Hessian and gradient.
  U = Uorig(:,latents);
  grad(latents) = grad(latents) + (w' .* -err) * U;
end
end

function covar = step_v_hessian(x_ordinal, w, Uorig, vorig, Hi, params, glmvars)

k = glmvars.k;
b = glmvars.b;
assert(cols(vorig) == k + b);
assert(cols(Uorig) == k + b);

if rows(Uorig) == 0
  covar = eye(k+b);
  return;
end

values = glmvars.values;

H = Hi;
D = eye(rows(Uorig));
didx = find(D);

for r = 1:b
  latents = [1:k k+r];
  x = x_ordinal;
  x(x < values(r)) = 1;
  x(x >= values(r)) = 0;

  % Compute coordinate scalines for the Hessian -- i.e., the weights in
  % iteratively reweighted least squares.
  uv = Uorig(:,latents) * vorig(:,latents)';
  raw_deriv = params.f_prime(uv);
  %D = spdiags(w .* raw_deriv, 0, rows(Uorig), rows(Uorig));
  D(didx) = w .* raw_deriv;

  % Add in the contribution of the Hessian and gradient.
  U = Uorig(:,latents);
  covar(latents, latents) = H(latents, latents) + U' * D * U;
end
end

% $$$ function [grad H] = step_v_grad_hessian(x_ordinal, w, Uorig, vorig, Hi, ...
% $$$                                         params, glmvars)
% $$$ 
% $$$ k = glmvars.k;
% $$$ b = glmvars.b;
% $$$ assert(cols(vorig) == k + b);
% $$$ assert(cols(Uorig) == k + b);
% $$$ 
% $$$ if rows(Uorig) == 0
% $$$   grad = zeros(1,k+b);
% $$$   H = eye(k+b);
% $$$   return;
% $$$ end
% $$$ 
% $$$ values = glmvars.values;
% $$$ 
% $$$ Bi = params.inv_h(vorig);
% $$$ Bi(k+1:end) = 0;
% $$$ 
% $$$ grad = Bi;
% $$$ H = Hi;
% $$$ 
% $$$ D = eye(rows(Uorig));
% $$$ didx = find(D);
% $$$ 
% $$$ for r = 1:b
% $$$   latents = [1:k k+r];
% $$$   x = x_ordinal;
% $$$   x(x < values(r)) = 1;
% $$$   x(x >= values(r)) = 0;
% $$$ 
% $$$   % Make predictions using all the latent factors and compute residuals
% $$$   uv = Uorig(:,latents) * vorig(:,latents)';
% $$$   expectx = params.f(uv);
% $$$   err = (x - expectx)';
% $$$   raw_deriv = params.f_prime(uv);
% $$$   %D = spdiags(w .* raw_deriv, 0, rows(Uorig), rows(Uorig));
% $$$   D(didx) = w .* raw_deriv;
% $$$ 
% $$$   % Add in the contribution of the Hessian and gradient.
% $$$   U = Uorig(:,latents);
% $$$   grad(latents) = grad(latents) + (w' .* -err) * U;
% $$$   H(latents, latents) = H(latents, latents) + U' * D * U;
% $$$ end
% $$$ end

function loss = xloss_natural_noreg(X, W, U, V, params, glmvars)

loss = 0;
b = glmvars.b;
k = glmvars.k;
values = glmvars.values;
for r = 1:b
  x_binary = full(X);
  x_binary(X < values(r)) = 1;
  x_binary(X >= values(r)) = 0;
  latents = [1:k k+r];
  T = U(:,latents) * V(:,latents)';
  loss = loss + full(params.loss_f(x_binary, W, T, params.f));
end
end

function obj = xloss_natural(X, W, U, V, params, glmvars)

loss = xloss_natural_noreg(X, W, U, V, params, glmvars);
reg_u = params.scale_g * sum(U(:).^2)/(2*numel(U));
reg_v = params.scale_h * sum(V(:).^2)/(2*numel(V));
obj = loss + reg_u + reg_v;
end

function obj = xloss(X, W, U, V, params, glmvars)

check_dims(X, rows(W), cols(W), 'X');
obj = xloss_natural(X, W, U, V, params, glmvars);
end

function obj = loss(Xt, U, V, params, glmvars)

obj = 0;

[n, m] = size(Xt);
for i = 1:m
  x = Xt(:,i)';
  w = zeros(size(x));
  w(x ~= 0) = 1/glmvars.numel;
  obj = obj + xloss_natural_noreg(x, w, U(i,:), V, params, glmvars);
end
obj = obj + params.scale_g * sum(U(:).^2)/(2*numel(U));
obj = obj + params.scale_h * sum(V(:).^2)/(2*numel(V));
end

function [p stepsize delta] = linesearch(f, p_init, eta, armijo, iter, params)

p = p_init;
delta = 0;
stepsize = NaN;
sufficient = false;

if params.linesearch && iter >= params.min_armijo_iter
  [p_new loss_old] = f(0);
  while (eta > params.mineta) && not(sufficient)
    [p_new loss_new] = f(eta);
    sufficient = armijo(loss_new, loss_old, eta);
    if sufficient
      p = p_new;
      stepsize = eta;
      delta = loss_new - loss_old;
      break;
    end
    eta = params.rho * eta;
  end
else
  p = f(eta);
  stepsize = eta;
end
end

function [step quit] = step_direction(grad, covar, inv_covar, params, glmvars)

threshold = eps;
quit = false;

if glmvars.gradient
  step = grad;
  quit = not(params.stochastic_opt) && norm(grad) <= threshold;
  return;
end

if not(params.stochastic_opt) && (grad * covar * grad')/2 <= threshold
  step = zeros(size(grad));
  quit = true;
end

step = grad * inv_covar;
end

% $$$ function [step quit] = step_direction(grad, covar, params, glmvars)
% $$$ 
% $$$ % TODO(ajit): This function does not respect many of the @glmparams
% $$$ % options regarding condition numbers and numerical stability, c.f.
% $$$ % check_covariance(). It's not a huge issue, since this implementation
% $$$ % is fast and stable.
% $$$ 
% $$$ if glmvars.clustering
% $$$   [step quit] = step_direction_clustering(grad, covar, params, glmvars);
% $$$   return;
% $$$ end
% $$$ 
% $$$ threshold = eps;
% $$$ 
% $$$ if glmvars.gradient
% $$$   % Requested gradient step
% $$$   step = grad;
% $$$   quit = not(params.stochastic_opt) && norm(grad) <= threshold;
% $$$   return;
% $$$ end
% $$$ 
% $$$ if not(params.stochastic_opt) && (grad * covar * grad')/2 <= threshold
% $$$   % Convergence condition for Newton step
% $$$   step = zeros(size(grad));
% $$$   quit = true;
% $$$   return;
% $$$ end
% $$$ 
% $$$ [R p] = chol(covar);
% $$$ if p ~= 0
% $$$   % Hessian not positive semi-definite. Replace with a gradient step
% $$$   step = grad;
% $$$   quit = false;
% $$$   return;
% $$$ end
% $$$ 
% $$$ % Vanilla Newton step direction
% $$$ Ri = inv(R);
% $$$ step = grad * Ri * Ri';
% $$$ quit = false;
% $$$ end
% $$$ 
% $$$ function [step quit] = step_direction_clustering(grad, covar, params, glmvars)
% $$$ 
% $$$ threshold = eps;
% $$$ 
% $$$ assert(not(glmvars.gradient), 'Cannot use gradient with clustering constr');
% $$$ assert(not(params.stochastic_opt), 'Stochastic Newton not supported yet');
% $$$ 
% $$$ if (grad * covar * grad')/2 <= threshold
% $$$   step = zeros(size(grad));
% $$$   quit = true;
% $$$   return;
% $$$ end
% $$$ 
% $$$ % Augment with the clustering constraint
% $$$ A = ones(rows(covar),1);
% $$$ covar = [ covar A ; A' 0 ];
% $$$ grad(end+1) = 0;
% $$$ covar
% $$$ grad
% $$$ stop
% $$$ 
% $$$ [R p] = chol(covar);
% $$$ if p ~= 0
% $$$   % If the Hessian is bogus quit
% $$$   fprintf('ack');
% $$$   step = zeros(size(grad));
% $$$   quit = true;
% $$$   return;
% $$$ end
% $$$ 
% $$$ % Invert and drop the Lagrange multiplier
% $$$ Ri = inv(R);
% $$$ step = grad * Ri * Ri';
% $$$ step = step(1:end-1);
% $$$ quit = false;
% $$$ end

%------------------------------------------------------------------------------
function check_matrix(X)

values = unique(X);
WarnIf(length(values) > 10, 'There are too many distinct values (%d)', ...
       length(values));
end
