function plot_compare_ranking(models, splits, alphas, desc, varargin)
% PLOT_COMPARE_RANKING - Plot the results of exp_compare_ranking
%
% plot_compare_ranking(models, splits, varargin)
%
% Arguments:
%
%
% Return Values:
%
% N/A

% Author: Ajit Singh (ajit@cs.cmu.edu)

error(nargchk(4, Inf, nargin));

check_type(models, 'cell');
check_type(splits, 'cell');
check_length(models, length(desc));
for i = 1:length(models)
  check_length(splits, rows(models{i}));
  check_length(alphas, cols(models{i}));
end

h1 = plot_method(models{1}, splits, alphas, 'glm');
hold on;
h2 = plot_method(models{2}, splits, alphas, 'glm');
h3 = plot_method(models{3}, splits, alphas, 'ch');

% Compute the ranking using popularity
fake = cell(length(splits), length(alphas));
[fake{:}] = deal(NaN);
h4 = plot_method(fake, splits, alphas, 'pop');
hold off;

% Make the plots look nice
fs = 14;
grid on;
set(gca, 'Box', 'off');
set(gca, 'FontSize', fs);
set(gca, 'Xtick', 0:.1:1);
xlim([0 1]);
xlabel('\alpha', 'FontSize', fs);
ylabel(name_scoring_method(), 'FontSize', fs);
legend({desc{:} 'Pop'});
legend boxoff;

set(h1, 'Marker', 'd', 'MarkerSize', 8.0, 'MarkerFaceColor', 'Black', ...
        'LineStyle', '-', 'LineWidth', 2.0, 'Color', 'Black');
set(h2, 'Marker', 'o', 'MarkerSize', 8.0, 'MarkerFaceColor', 'Red', ...
        'LineStyle', '-', 'LineWidth', 2.0, 'Color', 'Red');
set(h3, 'Marker', 's', 'MarkerSize', 8.0, 'MarkerFaceColor', 'Green', ...
        'LineStyle', '-', 'LineWidth', 2.0, 'Color', 'Green');
% $$$ set(h4, 'Marker', 'd', 'MarkerSize', 8.0, ...
% $$$         'LineStyle', '-', 'LineWidth', 2.0, 'Color', 'Blue');
set(h4, 'LineStyle', '--', 'LineWidth', 2.0);

function [h context] = plot_method(models, splits, alphas, type)

fprintf(1,'Plotting method\n');
scores = eval_method(models, splits, type);
y = mean(scores, 1);
yerr = std(scores, [ ], 1);
h = errorbar(alphas, y, yerr./2);

function rankings = glm_ranking(model, X)
% GLM_RANKING - Convert model3f to a ranking
%
% rankings = glm_rankings(model)
% Let e1 denote the entities indexed by the rows of model.U and e2 the
% entities indexed by the cols of model.V. This method generates a ranking
% of entities in e2 for each entity in e1.
%
% Arguments:
%
% model: Instance of @model3f.
%
% Return Values:
%
% rankings: Cell array of rankings, length equals rows(model.U).

% $$$ check_type(model, 'model3f');
U = model.U;
V = model.V;
rankings = cell(1, rows(U));
entities = 1:rows(V);
%pop = full(sum(X,1));
for i = 1:rows(U)
  rankings{i} = ranking(entities, U(i,:) * V');
end

function rankings = ch_ranking(model, X)
% CH_RANKING - Convert a pLSI-pHITS model to a ranking.
%
% rankings = ch_ranking(model)
% We want to rank documents for each each, p(d|t), where in the movie
% example d = movie, t = user. We know that p(d|t) \propto p(t|d)p(d)
% where p(t|d) = p(t|z)*p(z|d), ie, U*V' below. Need an estimate of
% the marginal probability of each document/movie.

U = model.TermFactors';
V = model.DocMixing';

%pop = ones(1, rows(V)); % really awful, p(t|d) and p(d|t) can differ a lot.
%pop = full(sum(X,1)); % works OK.
% $$$ warning('Need to estimate p(d) to get correct p(d|t), see notepad');
pop = estimate_document_probability(model, X);

rankings = cell(1, rows(U));
entities = 1:rows(V);
for i = 1:rows(U)
  rankings{i} = ranking(entities, U(i,:)*V' .* pop);
end

function rankings = pop_ranking(model, X);
% POP_RANKING - Generate a ranking in order of decreasing popularity

X = double(logical(X));
pop = full(sum(X,1));
entities = 1:cols(X);
rankings = cell(1, rows(X));
for i = 1:rows(X)
  rankings{i} = ranking(entities, pop);
end

function score = eval_ranking(rnk, entities)
% EVAL_RANKING - Evaluate the quality of a ranking
%
% Arguments:
%
% rnk: Instance of @ranking
% entities: The true held out entities (indices that are in rnk.rank)
%
% Return Values:
%
% score: The average rank of held out entities in the ranking

%score = eval_average_rank(rnk, entities);
score = eval_ap(rnk, entities, 'threshold', 200);

function str = name_scoring_method( )
%str = 'Average Rank';
str = 'Mean Average Precision';

function scores = eval_model(model, split, type)
% EVAL_GLM_MODEL - Evaluate a single GLM model using a ranking measure
%
% Arguments:
%
% model: Instance of @model3f.
% split: Test/training split (eg, @movie/data.m:test_train_splits) that
%        provides the held out entities.
%
% Return Values:
%
% scores: For each entity indexed by rows(model.U), ie, the rows of X,
%         we rank all the entities indexed by rows(model.V), ie, the cols
%         of X. The "true" held-out values used to evaluate the ranking
%         are the column indices of the held-out entries of X. scores
%         contains the ranking measure for each entity in rows(model.U)
%         which contained held-out links from X.

% $$$ if isa(model, 'model3f')
% $$$   rankings = glm_ranking(model, split.X_train);
% $$$ elseif isa(model, 'struct')
% $$$   rankings = ch_ranking(model, split.X_train);
% $$$ elseif isa(model, 'double') && not(isfinite(model))
% $$$   rankings = pop_ranking(model, split.X_train);
% $$$ else
% $$$   error('Bogus model');
% $$$ end
if strcmpi(type, 'glm')
  rankings = glm_ranking(model, split.X_train);
elseif strcmpi(type, 'ch')
  rankings = ch_ranking(model, split.X_train);
elseif strcmpi(type, 'pop')
  rankings = pop_ranking(model, split.X_train);
else
  error('Bogus model: %s\n', type);
end
scores = NaN * zeros(1, length(rankings));
for r = 1:length(rankings)
  entities = split.exj(find(split.exi == r));
  if not(isempty(entities))
    scores(r) = eval_ranking(rankings{r}, entities);
    %scores(r) = eval_average_rank(rankings{r}, entities);
  end
end
scores = scores(isfinite(scores));

function scores = eval_method(models, splits, type)
% EVAL_GLM_METHOD - Evaluate the quality of a GLM method on different splits
%                   and parameter settings.
%
% Arguments:
%
% models: s x p cell matrix, each cell containing a @model3f object.
% splits: length p cell array, each cell containing a split.
%
% Return Values:
%
% scores: s x p matrix whose entries correspond to the average of a ranking
%         measure (see eval_glm_model above).

check_length(splits, rows(models));
scores = zeros(rows(models), cols(models));
aggregate = @mean;
for s = 1:rows(models)
  for a = 1:cols(models)
    fprintf(1, 'split %d, params %d\n', s, a);
    per_user_scores = eval_model(models{s,a}, splits{s}, type);
    scores(s,a) = aggregate(per_user_scores);
  end
end
