function plot_stochastic(models, splits, params, obj, cputime, varargin)
% PLOT_STOCHASTIC - Plot the results of exp_stochastic
%
% plot_stochastic(models, splits, params, obj, cputime, varargin)
%
% Arguments:
%
% Let s be the number of splits and p be the number of methods
%
% obj: s x p cell matrix of training errors at each iteration
% cputime: s x p cell matrix of cpu-time (in sec) to finish first k iterations
% models: s x p cell matrix of models at each iteration
% params: 1 x p cell array of parameter settings. Assumes the only
%         differences are due to stochastic vs. Newton optimization.
% errorbars: Optional argument. If true use errorbars
%
% Return Values:
%
% N/A

% Author: Ajit Singh (ajit@cs.cmu.edu)

error(nargchk(5, Inf, nargin));

check_type(obj, 'cell');
check_type(cputime, 'cell');
check_type(params, 'cell');
check_dims(obj, rows(cputime), cols(cputime), 'obj');
check_length(params, cols(obj));
check_iterations(obj);
check_iterations(cputime);

[mode use_errorbars unused] = process_options(varargin, 'mode', [ ], ...
                                              'errorbars', false);
if isempty(mode)
  mode = 'training';
end
if use_errorbars && rows(obj) <= 1
  error('Opt arg "errorbars" cannot be used since there is only 1 split');
end

[loss.x loss.y] = evaluate_losses(models, splits, params);
[err.x err.y] = evaluate_errors(models, splits, params);

duration = average_columns(cputime);
names = describe_parameters(params);
if strcmpi(mode, 'training')
  [average std_dev] = average_columns(obj);
  plot_runs(duration, average, std_dev, names, ...
            'xlabel', 'CPU Time (s)', 'ylabel', 'Training Error', ...
            'usegrid', true, varargin{:});
elseif strcmpi(mode, 'test_loss_x')
  [average std_dev] = average_columns(loss.x);
  plot_runs(duration, average, std_dev, names, ...
            'xlabel', 'CPU Time (s)', 'ylabel', 'Test Loss -- X', ...
            'usegrid', true, varargin{:});
elseif strcmpi(mode, 'test_loss_y')
  [average std_dev] = average_columns(loss.y);
  plot_runs(duration, average, std_dev, names, ...
            'xlabel', 'CPU Time (s)', 'ylabel', 'Test Loss -- Y', ...
            'usegrid', true, varargin{:});
elseif strcmpi(mode, 'test_error_x')
  [average std_dev] = average_columns(err.x);
  plot_runs(duration, average, std_dev, names, ...
            'xlabel', 'CPU Time (s)', 'ylabel', 'Test Error -- X', ...
            'usegrid', true, varargin{:});
elseif strcmpi(mode, 'test_error_y')
  [average std_dev] = average_columns(err.y);
  plot_runs(duration, average, std_dev, names, ...
            'xlabel', 'CPU Time (s)', 'ylabel', 'Test Error -- Y', ...
            'usegrid', true, varargin{:});
end

function names = describe_parameters(params)
% DESCRIBE_PARAMETERS - Generate a string summarizing a parameterization.
%
% names = describe_parameters(params)
% Assumes that the only difference is stochatic or Newton optimization.

check_type(params, 'cell');
names = cell(1, length(params));
for i = 1:length(params)
  check_type(params{i}, 'glmparams');
  desc = '';
  if params{i}.stochastic_opt
    desc = sprintf('Stochastic Newton (batch = %d)', ...
                   params{i}.stochastic_batch_size);
  else
    desc = sprintf('Newton');
  end
  names{i} = desc;
end

function plot_runs(x, y, dev, names, varargin)
% PLOT_RUNS - Plot a set of x-y lineplots
%
% plot_runs(x, y, varargin)
% Plots a set of x-y curves, each corresponding to a pair of entries in
% cell arrays x and y. Each curve corresponds to a set of parameters
%
% Arguments:
%
% x: Cell array of x points (1 x p). Each cell a vector length k
% y: Cell array of y points (1 x p). Each cell a vector length k
% dev: Cell array of std deviations for the y points (1 x p). Cell vec len k.
%
% Return Values:
%
% N/A

check_type(x, 'cell');
check_type(y, 'cell');
check_type(dev, 'cell');
check_type(names, 'cell');
check_dims(x, rows(y), cols(y), 'x');
check_dims(x, rows(dev), cols(dev), 'x');
check_length(names, length(x));

[xlab ylab linewidth usegrid suptit use_errorbars unused] = ...
    process_options(varargin, 'xlabel', false, 'ylabel', false, ...
                    'linewidth', 2, 'usegrid', false, 'title', [ ], ...
                    'errorbars', false);
%linetypes = {'-', 'b--', 'g:', 'r-'};
linetypes = {'-k', '-r', '-g', '-b'};
linewidths = {3, 2, 2, 2};
figure;
hold on;
domain = [-Inf Inf];
range = [-Inf Inf];

for i = 1:length(x)
  lt = linetypes{mod(i-1, length(linetypes)) + 1};
  lw = linewidths{mod(i-1, length(linewidths)) + 1};
  if use_errorbars
    errorbar(x{i}, y{i}, dev{i}, lt, 'linewidth', linewidth);
  else
    plot(x{i}, y{i}, lt, 'linewidth', lw);
  end
  domain = update_plot_interval(x{i}, domain);
  range = update_plot_interval(y{i}, range);
end
xlim(domain);
ylim(range);
legend(names);
legend boxoff;
set(gca, 'XMinorTick', 'on', 'TickDir', 'out');
set(gca, 'YMinorTick', 'on');
set(gca, 'FontName', 'Helvetica', 'FontSize', 12);

if not(isempty(suptit))
  title(suptit);
end
if xlab
  xlabel(xlab);
end
if ylab
  ylabel(ylab);
end
if usegrid;
  grid;
end


