function [obj cputime models] = run_glm_params(splits, params, varargin)
% EXP_GLM_PARAMS - Compare 3-factor model on different splits and settings.
%
% [obj cputime models] = exp_glm_params(splits, params, varargin)
% For each split-params pair, run the three-factor model and output the
% results.
%
% Examples
%
% 1. To test a specific split and parameter sets
%
% -- split is a single split, param is a single params
% [obj cputime models] = exp_glm_params({split}, {param})
%
% Arguments
%
% splits: A cell array of splits (length s)
% params: Cell array of different optimization parameters (length p)
%
% Return Values:
%
% obj: s x p cell matrix where each cell is a length i vector of training
%      error at each iteration.
% cputime: s x p cell matrix where each cell is a length i vector where
%          the k^th element is the cpu time (in seconds) it took to complete
%          the first k iterations.
% models: s x p cell matrix where each cell is a length i cell array of
%         the model at each iteration.

% Author: Ajit Singh (ajit@cs.cmu.edu)

error(nargchk(2, Inf, nargin));

check_type(splits, 'cell');
check_type(params, 'cell');

obj = cell(length(splits), length(params));
cputime = cell(length(splits), length(params));
models = cell(length(splits), length(params));
for s = 1:length(splits)
  for p = 1:length(params)
    fprintf(1, ['============== Split %d/%d | Parameters %d/%d '...
                '==================\n'], ...
            s, length(splits), p, length(params));
    describe_glmparams(params{p});
    [model objective debug] = glm(splits{s}.X_train, splits{s}.wxf, ...
                                  splits{s}.wx_context, ...
                                  splits{s}.Y_train, splits{s}.wyf, ...
                                  splits{s}.wy_context, ...
                                  params{p}, splits{s}, varargin{:});
    obj{s,p} = objective;
    cputime{s,p} = debug.cputime;
    models{s,p} = debug.model;
  end
end
