function [metric cputime] = run_mmmf(splits, params, varargin)
% RUN_MMMF - Compare MMMF on different splits and settings
%
% [error cputime] = run_mmmf(splits, params, varargin)
% For each split-params pair, run max-margin matrix factorization and
% output the test error and training time.
%
% Examples
%
% 1. To test a specific split and parameter set
%
% -- split is a single split, param is a single param
% [error cputime] = run_mmmf({split}, {param});
%
% Arguments
%
% splits: A cell array of splits (length s)
% params: Cell array of different optimization parameters (length p)
%
% Return Values:
%
% error: s x p cell matrix where each cell is a scalar, the test error
% cputime: s x p cell matrix where each cell is a scalar, training time in sec.

% Author: Ajit Singh (ajit@cs.cmu.edu)

error(nargchk(2, Inf, nargin));

check_type(splits, 'cell');
check_type(params, 'cell');

metric = cell(length(splits), length(params));
cputime = cell(length(splits), length(params));

[type unused] = process_options(varargin, 'type', 'newton');
mock = false;
WarnIf(mock, 'Mocking the computation');

for p = 1:length(params)
  describe_mmmf_params(params{p}, type);
  for s = 1:length(splits)
    fprintf(1, ['============== MMMF (%s): Split %d/%d | Parameters %d/%d '...
                '==================\n'], ...
            type, s, length(splits), p, length(params));
    if mock
      err = rand;
      tim = rand;
    else
      [err tim] = dispatch(splits{s}, params{p}, type, varargin{:});
    end
    fprintf(1, 'Error (%s): %g Training Time: %g sec\n', ...
            func2str(splits{s}.eval_x), err, tim);
    metric{s,p} = err;
    cputime{s,p} = tim;
  end
end

function describe_mmmf_params(param, type)

if strcmpi(type, 'newton')
  describe_glmparams(param);
elseif strcmpi(type, 'conjgrad')
  describe_glmparams(param, 'print_opt', false, 'print_debug', false);
else
  error('RFA:InvalidArg', 'Ack %s', type);
end

function [err tim] = dispatch(split, param, type, varargin)

X = split.X_train;
if strcmpi(type, 'newton')
  [U V obj debug] = mmmf(X, param, varargin{:});
  xhat = generate_pred_mmmf(split.exi, split.exj, U, V, param.dims, param.f);
  maxval = max(max(X));
  minval = min(min(X));
  xhat(xhat > maxval) = maxval; % TODO(ajit): Data specific hack. Check if needd
  xhat(xhat < minval) = minval;
  err = split.eval_x(xhat, split.xtest);
  tim = debug.cputime(end);
elseif strcmpi(type, 'conjgrad')
  [Xpred tim] = wrap_fast_mmmf(X, param, varargin{:});
  xhat = Xpred(sub2ind(size(Xpred), split.exi, split.exj));
  err = split.eval_x(xhat, split.xtest);
else
  error('RFA:InvalidArg', 'Got %s, expected ''newton'' or ''conjgrad''', ...
        type);
end
