function prob = estimate_document_popularity(model, X)
% ESTIMATE_DOCUMENT_POPULARITY - Estimate p(d) in pLSI-pHITS
%
% prob = estimate_document_popularity(model, X)
% Computed using p(d) = p(d|z)p(z) where the quantities on the rhs are
% computed directed using the equations in the original pLSI paper, ie,
% we compute the E-statistics P(z|d,w) and use those to compute p(d|z) and p(z)
% This is basically 2/3 of the normal pLSI update--we don't compute p(w|z).
%
% Arguments:
%
% model: pLSI-pHITS model.
% X: training data.
%
% Return Values:
%
% prob = probability of each document in the pLSI-pHITS model.

K = rows(model.TermFactors);
Nw = cols(model.TermFactors);
Np = cols(model.DocMixing);
DocGivenLatent = NaN * zeros(K,Np); % p(d|z), each row a distribution
LatentProb = NaN * zeros(1,K); % p(z)

% The reason we do not normalize the columns of N here is p(d) is fit to
% match the empirical probability of each document, roughly,
% sum(N,1)./sum(sum(N)). If N is column normalized the distribution will
% be very close to uniform, and so will the output prob.
N = X;
R = full(sum(sum(N)));
[Ni,Nj] = find(N);

% p(zk|ti,dj) -- each column is a distribution
LatentByDocs = model.TermFactors(:,Ni) .* model.DocMixing(:,Nj);
LatentByDocs = LatentByDocs.^model.beta;
LatentByDocs = safe_normalize_cols(LatentByDocs);

for k = 1:K
  X = N .* sparse(Ni,Nj,LatentByDocs(k,:),Nw,Np);
  Z = full(sum(sum(X)));
  LatentProb(k) = (1/R) * Z;
  DocGivenLatent(k,:) = sum(X,1)./Z;
end

prob = LatentProb * DocGivenLatent;
prob = prob./sum(prob); % fixes precision issues
