function model = random_model4f(X, Y, Q, params)
% RANDOM_MODEL4F - Create a random set of parameters for the 3-factor model
%
% There are two extra columns of U, which when paired with the first
% two extra columns of V constitute bias terms for each row and column of X
% The last two extra columns of V, along with the extra columns of Z
% are bias terms for each row and column of Y.
%
% Note: All the biases are computed in the space of natural parameters,
% since we're pushing the bias terms into U, V, and Z. This is not the
% same as X = f(UV') + Bias, which would be a bias computed in the
% expectation space. Since f(a + b) != f(a) + f(b) for most links
% you cannot easily interpret the bias.
%
% model = random_model3f(X, Y, k)
% Creates an instance of @model3f whose factors are drawn uniformly at
% random. Useful for creating a shared seed for different runs of glm.
%
% This is not a member of @model3f because matlab does not support static
% member functions for classes.
%
% Arguments:
%
% X: Data matrix (m x n)
% Y: Data matrix (n x r)
%
% Return Values:
%
% model: Random model. Instance of @glmparams.

% Author: Ajit Singh (ajit@cs.cmu.edu)

error(nargchk(4, 4, nargin));
%check_type(params, 'glmparams');

k = params.dims;
assert(k > 0 & k == fix(k), 'k is not an embedding dimension (%f)', k);
assert(cols(X) == rows(Y), ...
       'X and Y do not conform (%d x %d) and (%d x %d)', size(X), size(Y));

[m n] = size(X);
[ig r] = size(Y);
[ig s] = size(Q);
U = normalize_rows(rand(m, k));
V = normalize_rows(rand(n, k));
Z = normalize_rows(rand(r, k));
R = normalize_rows(rand(s, k));
if params.add_bias_term
  error('Bias term not supported on 4f');
% $$$   U = [U  ones(m, 1)   rand(m, 1)   zeros(m, 1)  zeros(m, 1)];
% $$$   V = [V  rand(n, 1)   ones(n, 1)   ones(n, 1)   rand(n, 1)];
% $$$   Z = [Z  zeros(r, 1)  zeros(r, 1)  rand(r, 1)   ones(r, 1)];
end
model = model4f(U, V, Z, R, params);
